"""Base exceptions module"""


class RPCError(Exception):
    """Base error class for RPC protocols"""

    message: str = None

    def __init__(self, *args):
        super().__init__(*args)
        if len(args) > 0:
            self.message = str(args[0])

    def __str__(self) -> str:
        if self.message:
            return f"{self.__class__.__name__} ({self.message})"
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        if self.message:
            return f"<{self.__class__.__name__} ({self.message})>"
        return f"<{self.__class__.__name__}>"

    def as_dict(self) -> dict[str, str]:
        return {
            "error": self.__class__.__name__,
            "message": self.message or "",
        }


class SerializationError(RPCError):
    """Serialization error"""


class DeserializationError(RPCError):
    """Deserialization error"""


class InternalServerError(RPCError):
    """Server error"""


class BadRequestError(RPCError):
    """Bad request error"""


class NotFoundError(BadRequestError):
    """Not found error"""


class InvalidParametersError(BadRequestError):
    """Invalid parameters error"""


class AuthenticationFailedError(RPCError):
    """Authentication failed"""


class InvalidTokenError(AuthenticationFailedError):
    """Invalid token error"""


class ForbiddenError(AuthenticationFailedError):
    """Forbidden error"""


class InvalidPermissionsError(BadRequestError):
    """Invalid permissions error"""



import json
from collections.abc import Callable
from functools import partial
from io import RawIOBase
from typing import Any
from typing import Optional
from typing import Type
from typing import Union

try:
    import msgpack
except ImportError:
    msgpack = None

try:
    import orjson
except ImportError:
    orjson = None

try:
    import ujson
except ImportError:
    ujson = None

from drakaina.exceptions import DeserializationError
from drakaina.exceptions import SerializationError


class BaseSerializer:
    content_type: str

    def serialize(self, obj: Any) -> bytes:
        raise NotImplementedError

    def deserialize(self, s: Union[bytes, str]) -> Any:
        raise NotImplementedError


class BaseJsonSerializer(BaseSerializer):
    dumps: Callable
    loads: Callable
    DecodeError: Type[Exception]
    content_type: str = "application/json"

    def serialize(self, obj: Any) -> bytes:
        try:
            return self.dumps(obj=obj).encode(encoding="utf-8")
        except (RecursionError, TypeError, ValueError) as exc:
            raise SerializationError from exc

    def deserialize(self, s: Union[bytes, str]) -> Any:
        try:
            return self.loads(s)
        except (self.DecodeError, ValueError) as exc:
            raise DeserializationError from exc


class JsonSerializer(BaseJsonSerializer):
    def __init__(
        self,
        encode_kw: Optional[dict] = None,
        decode_kw: Optional[dict] = None,
        encoder: Optional[Type[json.JSONEncoder]] = None,
        decoder: Optional[Type[json.JSONDecoder]] = None,
    ):
        super().__init__()
        self.dumps = partial(
            json.dumps,
            cls=encoder,
            ensure_ascii=False,
            separators=(",", ":"),
            **encode_kw or {},
        )
        self.loads = partial(json.loads, cls=decoder, **decode_kw or {})
        self.DecodeError = json.JSONDecodeError


class ORJsonSerializer(BaseJsonSerializer):
    def __init__(
        self,
        encode_kw: Optional[dict] = None,
        decode_kw: Optional[dict] = None,
    ):
        if orjson is None:
            raise ModuleNotFoundError(
                "To use `ORJsonSerializer` you must install "
                "the `orjson` package",
            )

        super().__init__()
        self.dumps = partial(orjson.dumps, **encode_kw or {})
        self.loads = partial(orjson.loads, **decode_kw or {})
        self.DecodeError = orjson.JSONDecodeError

    def serialize(self, obj: Any) -> bytes:
        try:
            return self.dumps(obj)
        except (
            orjson.JSONEncodeError,
            RecursionError,
            TypeError,
            ValueError,
        ) as exc:
            raise SerializationError from exc


class UJsonSerializer(BaseJsonSerializer):
    def __init__(
        self,
        encode_kw: Optional[dict] = None,
        decode_kw: Optional[dict] = None,
    ):
        if ujson is None:
            raise ModuleNotFoundError(
                "To use `UJsonSerializer` you must install the `ujson` package",
            )

        super().__init__()
        self.dumps = partial(ujson.dumps, ensure_ascii=False, **encode_kw or {})
        self.loads = partial(ujson.loads, **decode_kw or {})
        self.DecodeError = ujson.JSONDecodeError


class MSGPackSerializer(BaseSerializer):
    content_type: str = "application/x-msgpack"

    def __init__(
        self,
        packer: Optional[Type["msgpack.Packer"]] = None,
        unpacker: Optional[Type["msgpack.Unpacker"]] = None,
        pack_kw: Optional[dict] = None,
        unpack_kw: Optional[dict] = None,
    ):
        if msgpack is None:
            raise ModuleNotFoundError(
                "To use `MSGPackSerializer` you must install "
                "the `msgpack` package",
            )

        self.packer = packer or msgpack.Packer
        self.unpacker = unpacker
        self.pack_kw = pack_kw or {}
        self.unpack_kw = unpack_kw or {}

    def serialize(self, obj: Any) -> bytes:
        try:
            return self.packer(**self.pack_kw).pack(obj)
        except (OverflowError, TypeError, ValueError) as exc:
            raise SerializationError from exc

    def deserialize(self, s: Union[bytes, str, RawIOBase]) -> Any:
        try:
            if self.unpacker is not None:
                return self.unpacker(s, **self.unpack_kw)
            return msgpack.unpackb(s, **self.unpack_kw)
        except (
            msgpack.UnpackException,
            msgpack.ExtraData,
            TypeError,
            ValueError,
        ) as exc:
            raise DeserializationError from exc



from __future__ import annotations

import re
from typing import Iterable

from drakaina import ENV_AUTH_PAYLOAD
from drakaina import ENV_AUTH_SCOPES
from drakaina import ENV_IS_AUTHENTICATED
from drakaina import ENV_USER
from drakaina import ENV_USER_ID
from drakaina.contrib.jwt import SUPPORTED_ALGORITHMS
from drakaina.contrib.jwt.errors import InvalidJWTTokenError
from drakaina.contrib.jwt.types import RevokeChecker
from drakaina.contrib.jwt.types import ScopesGetter
from drakaina.contrib.jwt.types import TokenGetter
from drakaina.contrib.jwt.types import UserGetter
from drakaina.contrib.jwt.utils import decode_jwt_token
from drakaina.exceptions import AuthenticationFailedError
from drakaina.exceptions import ForbiddenError
from drakaina.middleware.base import BaseMiddleware
from drakaina.types import ASGIReceive
from drakaina.types import ASGIScope
from drakaina.types import ASGISend
from drakaina.types import WSGIApplication
from drakaina.types import WSGIEnvironment
from drakaina.types import WSGIResponse
from drakaina.types import WSGIStartResponse
from drakaina.utils import get_cookies
from drakaina.utils import iterable_str_arg


class JWTAuthenticationMiddleware(BaseMiddleware):
    """The middleware supporting JWT tokens.

    :param app: WSGI Application.
    :param algorithms: List of supported algorithms.
    :param secret_key: A secret key for signature verification.
    :param public_key: Public key for signature verification in
        asymmetric algorithms.
    :param credentials_required: If `True`, the request is expected to
        contain credential data. If they are missing, an exception is raised.
        Default: `True`.
    :param prefix: The token prefix in the authorization headers.
        Default: `"Bearer"`.
    :param use_cookies: If true, the token will be searched in the cookie
        headers for the specified key. Default: `False`.
    :param cookie_key: The key by which the token is stored in the cookie
        headers.
    :param decode_options: The dictionary is passed as is as the `options`
        argument to the `pyjwt.decode` function. Default: `{
            "require": [],  # Require claims
            "verify_iss": False,  # Issuer
            "verify_aud": False,  # Audience
            "verify_exp": True,   # Expiry
            "verify_nbf": False,  # Not Before
            "verify_iat": True,   # Issued at
            "verify_jti": False,  # JWT ID
        }`
    :param verify_values: A dictionary of key-value pairs, where key is
        the name of the key in the token payload, and the value to check can be
        a specific value, a list of possible values, or a re.Pattern object
        (the result of re.compile).
    :param leeway_value: The value of leeway in seconds. Default: `0`.
    :param user_id_field: The name of the key in the token payload to store
        the user ID. Default: `"user_id"`.
    :param token_getter: Callable for an alternative way to get a token.
    :param user_getter: Callable to retrieve a user object.
    :param scopes_getter: Callable to get permission scopes.
    :param revoke_checker: Callable to verify that the token is revoked.
    :param kwargs: Other arguments to pass to the constructor of the base class.

    """

    __slots__ = (
        "_algorithms",
        "__verify_key",
        "_credentials_required",
        "_prefix",
        "_cookie_key",
        "_decode_options",
        "_verify_values",
        "_leeway",
        "_user_id_field",
        "get_token",
        "get_user",
        "is_revoked",
        "get_scopes",
    )

    def __init__(
        self,
        app: WSGIApplication,
        algorithms: str | Iterable[str] = SUPPORTED_ALGORITHMS,
        secret_key: str = None,
        public_key: str = None,
        credentials_required: bool = True,
        prefix: str = "Bearer",
        use_cookies: bool = False,
        cookie_key: str = "jwt",
        decode_options: dict[str, bool | list[str]] | None = None,
        verify_values: dict[str, str | Iterable[str] | re.Pattern]
        | None = None,
        leeway_value: int | float = 0,
        user_id_field: str = "user_id",
        token_getter: TokenGetter | None = None,
        user_getter: UserGetter | None = None,
        scopes_getter: ScopesGetter | None = None,
        revoke_checker: RevokeChecker | None = None,
        **kwargs,
    ):
        super().__init__(app, **kwargs)

        self._algorithms = iterable_str_arg(algorithms)
        self.__verify_key = secret_key or public_key
        self._credentials_required = credentials_required
        self._prefix = prefix
        self._cookie_key = cookie_key
        self._decode_options = (
            decode_options
            if decode_options is not None
            else {
                "require": [],
                "verify_iss": False,
                "verify_aud": False,
                "verify_exp": True,
                "verify_nbf": False,
                "verify_iat": True,
                "verify_jti": False,
            }
        )
        self._verify_values = verify_values if verify_values is not None else {}
        self._leeway = leeway_value
        self._user_id_field = user_id_field

        if callable(token_getter):
            self.get_token = token_getter
        elif use_cookies:
            self.get_token = self._token_from_cookies
        else:
            self.get_token = self._token_from_auth_header

        self.is_revoked = revoke_checker
        self.get_user = user_getter
        self.get_scopes = scopes_getter

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        if environ["REQUEST_METHOD"] == "OPTIONS":
            return self.app(environ, start_response)

        token = self.get_token(environ)

        if not token and self._credentials_required:
            raise AuthenticationFailedError("Credential required")

        if token is not None:
            try:
                token_payload = decode_jwt_token(
                    token=token,
                    algorithms=self._algorithms,
                    verify_key=self.__verify_key,
                    verify=True,
                    decode_options=self._decode_options,
                    verify_values=self._verify_values,
                    leeway=self._leeway,
                )
                token_is_valid = True
            except InvalidJWTTokenError as error:
                if self._credentials_required:
                    raise error
                else:
                    token_payload = decode_jwt_token(
                        token=token,
                        algorithms=self._algorithms,
                        verify_key=self.__verify_key,
                        verify=False,
                        decode_options=None,
                        verify_values=self._verify_values,
                        leeway=self._leeway,
                    )
                    token_is_valid = environ.get(ENV_IS_AUTHENTICATED, False)

            if callable(self.is_revoked):
                if self.is_revoked(environ, token_payload):
                    raise ForbiddenError("Token is revoked")

            environ[ENV_USER_ID] = token_payload.get(self._user_id_field)
            environ[ENV_IS_AUTHENTICATED] = token_is_valid
            environ[ENV_AUTH_PAYLOAD] = token_payload

            if callable(self.get_user):
                environ[ENV_USER] = self.get_user(environ, token_payload)
            if callable(self.get_scopes):
                environ[ENV_AUTH_SCOPES] = self.get_scopes(
                    environ,
                    token_payload,
                )

        return self.app(environ, start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        await self.app(scope, receive, send)

    def _token_from_auth_header(
        self,
        request: ASGIScope | WSGIEnvironment,
    ) -> str | None:
        auth_header = request.get("HTTP_AUTHORIZATION")

        if auth_header is None:
            return None

        try:
            parts = auth_header.strip().split(" ")
        except ValueError:
            raise AuthenticationFailedError("Invalid `Authorization` header")

        if len(parts) == 0 or parts[0] not in self._prefix:
            return None

        if len(parts) != 2:
            raise InvalidJWTTokenError(
                "The `Authorization` header must contain two values "
                "separated by a space",
            )

        return parts[1]

    def _token_from_cookies(
        self,
        request: ASGIScope | WSGIEnvironment,
    ) -> str | None:
        cookie_header = request.get("HTTP_COOKIE")

        if cookie_header is None:
            return None

        cookies = get_cookies(cookie_header)

        return cookies.get(self._cookie_key)



from __future__ import annotations

import re
from datetime import datetime
from datetime import timedelta

try:
    from datetime import UTC
except ImportError:
    from datetime import timezone

    UTC = timezone(timedelta(0), "UTC")

from typing import Any
from typing import Iterable
from typing import MutableMapping
from uuid import uuid4

from jwt import InvalidAlgorithmError
from jwt import InvalidTokenError
from jwt import PyJWKClient
from jwt.algorithms import has_crypto
from jwt.algorithms import requires_cryptography
from jwt.api_jwt import decode_complete
from jwt.api_jwt import encode

from drakaina.contrib.jwt import RESERVED_FIELDS
from drakaina.contrib.jwt import SUPPORTED_ALGORITHMS
from drakaina.contrib.jwt.errors import InvalidJWTTokenError
from drakaina.contrib.jwt.errors import JWTBackendError
from drakaina.contrib.jwt.errors import ValidationJWTTokenError
from drakaina.utils import iterable_str_arg


def datetime_utc_now() -> datetime:
    """Returns the current date/time in the UTC time zone.

    :return: The current date/time is in the UTC time zone.

    """
    return datetime.now(tz=UTC)


def datetime_to_timestamp(dt: datetime) -> int:
    """Converts the specified date/time to unixtime timestamp.

    :param dt: Date/time for conversion.
    :return: Unixtime timestamp.

    """
    return int(dt.timestamp())


def datetime_from_timestamp(ts: int) -> datetime:
    """Converts the specified unixtime timestamp to the python date/time type.

    :param ts: Timestamp for conversion.
    :return: Date/time from timestamp.

    """
    return datetime.fromtimestamp(ts, tz=UTC)


# Claims of specification JWT


def set_issuer(payload: dict[str, Any], value: str, claim: str = "iss"):
    """Sets the issuer of the token.

    https://tools.ietf.org/html/rfc7519#section-4.1.1

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_subject(payload: dict[str, Any], value: str, claim: str = "sub"):
    """Sets the subject of the token.

    https://tools.ietf.org/html/rfc7519#section-4.1.2

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_audience(payload: dict[str, Any], value: str, claim: str = "aud"):
    """Sets the audience of the token.

    https://tools.ietf.org/html/rfc7519#section-4.1.3

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_expiration(
    payload: dict[str, Any],
    expiration_time: datetime | int,
    claim: str = "exp",
):
    """Sets the expiration time of a token.

    https://tools.ietf.org/html/rfc7519#section-4.1.4

    :param payload: Token payload.
    :param expiration_time: Date/time value to set the claim.
    :param claim: Claim short name (key).

    """
    if isinstance(expiration_time, (int, float)):
        payload[claim] = int(expiration_time)
    else:
        payload[claim] = datetime_to_timestamp(expiration_time)


def set_not_before(
    payload: dict[str, Any],
    activation_time: datetime | int,
    claim: str = "nbf",
):
    """Sets the time `Not before`.

    https://tools.ietf.org/html/rfc7519#section-4.1.5

    :param payload: Token payload.
    :param activation_time: Date/time value to set the claim.
    :param claim: Claim short name (key).

    """
    if isinstance(activation_time, (int, float)):
        payload[claim] = int(activation_time)
    else:
        payload[claim] = datetime_to_timestamp(activation_time)


def set_issued_at(
    payload: dict[str, Any],
    now: datetime | int,
    claim: str = "iat",
):
    """Sets the time of token issuance.

    https://tools.ietf.org/html/rfc7519#section-4.1.6

    :param payload: Token payload.
    :param now: Current date/time value to set the claim.
    :param claim: Claim short name (key).

    """
    if isinstance(now, (int, float)):
        payload[claim] = int(now)
    else:
        payload[claim] = datetime_to_timestamp(now)


def set_token_id(
    payload: dict[str, Any],
    value: str | bool | None = None,
    claim: str = "jti",
):
    """Sets the token identifier.

    https://tools.ietf.org/html/rfc7519#section-4.1.7

    :param payload: Token payload.
    :param value: Value to set the claim. Default UUID4.
    :param claim: Claim short name (key).

    """
    if value is True or value is None:
        value = uuid4().hex
    payload[claim] = value


def check_expiration(
    payload: dict[str, Any],
    now: datetime,
    claim: str = "exp",
):
    """Checks if the timestamp value has passed in the given claim
    (since the given time value in `now`).
    If yes, it causes a `InvalidTokenError` with an error message.

    :param payload: Token payload.
    :param now: Current date/time value to check the claim.
    :param claim: Claim short name (key).

    :raise InvalidJWTTokenError:
    :raise ValidationJWTTokenError:

    """
    try:
        expiration_value = payload[claim]
    except KeyError:
        raise InvalidJWTTokenError(f"Token has no `{claim}` claim")

    expiration = datetime_from_timestamp(expiration_value)
    if now >= expiration:
        raise ValidationJWTTokenError(f"Token `{claim}` claim has expired")


def check_not_before(
    payload: dict[str, Any],
    now: datetime,
    claim: str = "nbf",
):
    """Checks if the timestamp value has reached the specified claim
    (the moment of the specified `now` time value).
    If so, raises a `InvalidTokenError` with an error message.

    :param payload: Token payload.
    :param now: Current date/time value to check the claim.
    :param claim: Claim short name (key).

    :raise InvalidJWTTokenError:
    :raise ValidationJWTTokenError:

    """
    try:
        not_before_value = payload[claim]
    except KeyError:
        raise InvalidJWTTokenError(f"Token has no `{claim}` claim")

    not_before_value = datetime_from_timestamp(not_before_value)
    if now < not_before_value:
        raise ValidationJWTTokenError(f"Token `{claim}` claim not reached")


# Custom JWT claims


def set_token_type(payload: dict[str, Any], value: str, claim: str = "jtt"):
    """Sets the token type claim.

    :param payload: Token payload.
    :param value: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = value


def set_permission_scopes(
    payload: dict[str, Any],
    scopes: str | Iterable[str],
    claim: str = "scp",
):
    """Sets the claim of the permissions scopes.

    :param payload: Token payload.
    :param scopes: Value to set the claim.
    :param claim: Claim short name (key).

    """
    payload[claim] = ",".join(iterable_str_arg(scopes))


# Token handlers


def decode_jwt_token(
    token: str,
    verify_key: str,
    algorithms: Iterable[str] = SUPPORTED_ALGORITHMS,
    verify: bool = True,
    decode_options: dict[str, bool | list[str]] | None = None,
    verify_values: dict[str, str | Iterable[str] | re.Pattern] | None = None,
    leeway: int | float = 0,
    jwk_url: str | None = None,
) -> dict[str, Any]:
    """Performs validation of the provided token and returns
    its payload dictionary.

    :param token: JWT token.
    :param verify_key: The token verification key.
        May be a signature key in symmetric algorithms, or a public key
        in asymmetric signature algorithms.
    :param algorithms: Supported algorithms.
    :param verify: Verify token.
    :param decode_options: Options argument for `pyjwt.decode`.
        Default PyJWT values:
        `{
            "verify_signature": True,
            "verify_exp": True,
            "verify_nbf": True,
            "verify_iat": True,
            "verify_aud": True,
            "verify_iss": True,
            "require": [],
        }`
    :param verify_values: Values to be verified.
    :param leeway: Leeway time in seconds.
    :param jwk_url: URL for the PyJWK client.
        It gets the validation key for the specified jwt-token.
    :return: Decoded JWT token payload.

    :raise JWTBackendError: In the case of incorrectly passed
        parameters and arguments.
    :raise ValidationJWTTokenError: In case the token has an invalid signature
        or invalid payload data.

    """

    if jwk_url:
        jwks_client = PyJWKClient(jwk_url)
        verify_key = jwks_client.get_signing_key_from_jwt(token).key

    verify_values = {} if verify_values is None else verify_values.copy()
    options = {"verify_signature": verify, **(decode_options or {})}
    params = {"leeway": leeway}

    # Provide values to check in the pyjwt module
    if "iss" in verify_values and not isinstance(
        verify_values["iss"],
        (Iterable, re.Pattern),  # for pyjwt issuer must be single
    ):
        params["issuer"] = verify_values.pop("iss", None)
    if "aud" in verify_values and not isinstance(
        verify_values["aud"],
        re.Pattern,  # for pyjwt audience must be single or iterable
    ):
        params["audience"] = verify_values.pop("aud")

    try:
        token_data = decode_complete(
            jwt=token,
            key=verify_key,
            algorithms=algorithms,
            options=options,
            **params,
        )
    except InvalidAlgorithmError as error:
        raise InvalidJWTTokenError("Invalid algorithm specified") from error
    except InvalidTokenError:
        raise InvalidJWTTokenError("Token is invalid or expired")

    if verify:
        _extra_validation(token_data, verify_values)

    return token_data["payload"]


def encode_jwt_token(
    signing_key: str,
    algorithm: str = "HS256",
    payload: dict[str, Any] | None = None,
    issuer: str | None = None,
    subject: str | None = None,
    audience: str | None = None,
    expiration: datetime | timedelta | int | None = None,
    not_before: datetime | timedelta | int | bool | None = None,
    issued_at: datetime | int | bool | None = None,
    token_id: int | str | bool | None = None,
    token_type: str | None = None,
    permission_scopes: str | Iterable[str] | None = None,
    headers: dict[str, Any] | None = None,
) -> str:
    """Returns an encoded token with the specified payload and
    specified parameters.

    The claims from the specification for JWT tokens are expected in
    the corresponding key parameters.

    :param signing_key: The secret key to sign a token.
    :param algorithm: Token Signing Algorithm.
    :param payload: Token payload.
    :param issuer: Issuer.
    :param subject: Subject.
    :param audience: Audience.
    :param expiration: Expiration time.
    :param not_before: Not before time.
        If `True`, the current time is applied.
    :param issued_at: Issued at the time.
        If `True`, the current time is applied. Default: `True`.
    :param token_id: Token ID.
        If `True`, then uuid4 is generated.
    :param token_type: The type of token, if you need it.
        It has the standard claim `jtt'.
    :param permission_scopes: Scope of permits, if you need them.
        It has the standard claim `scp'.
    :param headers: JWT token headers. RFC7519#section5
    :return: Encoded JWT token.

    """
    _validate_algorithm(algorithm)

    if not isinstance(payload, (dict, MutableMapping)):
        payload = {}
    else:
        payload = payload.copy()
        _validate_payload(payload)

    if issuer:
        set_issuer(payload, issuer)
    if subject:
        set_subject(payload, subject)
    if audience:
        set_audience(payload, audience)

    now = datetime_utc_now()

    if expiration:
        if isinstance(expiration, timedelta):
            expiration = now + expiration
        set_expiration(payload, expiration)
    if not_before:
        if isinstance(not_before, timedelta):
            not_before = now + not_before
        elif not_before is True:
            not_before = now
        set_not_before(payload, not_before)
    if issued_at:
        if issued_at is True:
            issued_at = now
        set_issued_at(payload, issued_at)

    if token_id:
        set_token_id(payload, token_id)

    if token_type:
        set_token_type(payload, token_type)
    if permission_scopes:
        set_permission_scopes(payload, permission_scopes)

    return encode(
        payload=payload,
        key=signing_key,
        algorithm=algorithm,
        headers=headers,
        json_encoder=None,  # todo: rewrite for use custom JSON serializer
    )


# Helpful functions


def copy_payload(
    payload: dict[str, Any],
    no_copy: Iterable[str] = ("exp", "nbf", "iat", "jti", "jtt", "scp"),
) -> tuple[dict[str, Any], ...]:
    """This function helps to copy the data of the old token and prepare
    the payload and parameters of the new token creation function.

    The data that can be safely copied will be returned by the first mapping.
    And the data of the JWT specification fields expected by
    the `encode_jwt_token` function as function parameters will be
    returned by the second mapping.

    This can be applied e.g. when updating an old token, keeping
    permanent data, and updating data critical for token verification
    and rotation.

    Example ::

        >>> payload, params = copy_payload(old_token_payload)
        >>> token = encode_jwt_token("_secret_", payload=payload, **params)

    :param payload: The payload for copying.
    :param no_copy: Claims that don't require copying.
    :return: A tuple of two elements, where the first is a filtered copy
        of the payload and the second is a mapping with parameters
        to pass to the `encode_jwt_token` function.

    """
    payload = payload.copy()

    # Mapping as { claim_abbr: param_name }
    encode_params_map = {
        "iss": "issuer",
        "sub": "subject",
        "aud": "audience",
        "exp": "expiration",
        "nbf": "not_before",
        "iat": "issued_at",
        "jti": "token_id",
        "jtt": "token_type",
        "scp": "permission_scopes",
    }
    # Preparing parameters for the `encode_jwt_token` function
    new_encode_params = {}
    for claim, param_name in encode_params_map.items():
        if claim in payload and claim not in no_copy:
            new_encode_params[param_name] = payload.pop(claim)

    # Copying the remaining payload
    new_payload = {
        claim: value for claim, value in payload.items() if claim not in no_copy
    }

    return new_payload, new_encode_params


def _extra_validation(
    decoded_token: dict[str, dict[str, Any] | str],
    verify_values: dict[str, str | Iterable[str] | re.Pattern | dict] = None,
):
    """Performs an additional check that is not performed by the pyjwt module.

    The check is performed by comparing the data provided for the check with
    the data included in the token payload with the same keys.

    :param decoded_token: Dict returned from `pyjwt.decode_complete`. Must
        contain a `header' and `payload'.
    :param verify_values: Dict with data to be validated. The keys must match
        the dictionary fields with the token payload.
    :raise ValidationJWTTokenError: In case the token has an invalid
        payload or header data.

    """
    if verify_values is None:
        verify_values = {}

    header = decoded_token["header"]
    verify_header_values = verify_values.pop("header", {})
    try:
        for key, verify_value in verify_header_values.items():
            token_value = header.get(key)
            if isinstance(verify_value, (str, int)):
                assert token_value == verify_value
            elif isinstance(verify_value, (list, set, tuple)):
                assert token_value in verify_value
            elif isinstance(verify_value, re.Pattern):
                assert re.match(verify_value, token_value)
    except AssertionError as error:
        raise ValidationJWTTokenError("Invalid token header") from error

    payload = decoded_token["payload"]
    try:
        for key, verify_value in verify_values.items():
            token_value = payload.get(key)
            if isinstance(verify_value, (str, int)):
                assert token_value == verify_value
            elif isinstance(verify_value, (list, set, tuple)):
                assert token_value in verify_value
            elif isinstance(verify_value, re.Pattern):
                assert verify_value.match(token_value)
    except AssertionError as error:
        raise ValidationJWTTokenError("Invalid token payload") from error


def _validate_algorithm(algorithm: str):
    """Check that the specified algorithm is recognized, and that
    `cryptography' is installed for those algorithms that require it.

    :param algorithm: Algorithm to verify.
    :raise JWTBackendError: If the algorithm is not recognized or not supported.
    :raise ModuleNotFoundError: If the `cryptography' package is not installed.

    """
    if algorithm not in SUPPORTED_ALGORITHMS:
        raise JWTBackendError(f"Unrecognized algorithm type '{algorithm}'")

    if algorithm in requires_cryptography and not has_crypto:
        raise ModuleNotFoundError(
            f"To use `{algorithm}` you must install the `cryptography` package",
        )


def _validate_payload(payload: dict[str, Any]):
    """Checks that the payload does not use reserved fields.

    :param payload: Payload to verify.
    :raise JWTBackendError: If the provided payload has reserved fields.

    """
    for reserved_jwt_claim in RESERVED_FIELDS:
        if reserved_jwt_claim in payload:
            if reserved_jwt_claim == "jtt":
                raise JWTBackendError(
                    f"The claim `{reserved_jwt_claim}` is reserved in "
                    f"Drakaina to be specified as `token_type`.",
                )
            elif reserved_jwt_claim == "scp":
                raise JWTBackendError(
                    f"The claim `{reserved_jwt_claim}` is reserved in "
                    f"Drakaina to be specified as `permission_scopes`.",
                )
            raise JWTBackendError(
                "Please use keyword arguments to fill in "
                "the JWT specification fields.",
            )



from __future__ import annotations

from logging import Logger

from drakaina.middleware.base import BaseMiddleware
from drakaina.rpc_protocols import BaseRPCProtocol
from drakaina.types import ASGIApplication
from drakaina.types import ASGIReceive
from drakaina.types import ASGIScope
from drakaina.types import ASGISend
from drakaina.types import WSGIApplication
from drakaina.types import WSGIEnvironment
from drakaina.types import WSGIResponse
from drakaina.types import WSGIStartResponse


class ExceptionMiddleware(BaseMiddleware):
    """The middleware for handling unhandled exceptions in the application
    according to the RPC protocol.

    """

    __slots__ = ("handler", "_logger")

    def __init__(
        self,
        app: ASGIApplication | WSGIApplication,
        handler: BaseRPCProtocol,
        logger: Logger | None = None,
        is_async: bool = False,
    ):
        super().__init__(app=app, is_async=is_async)

        self.handler = handler
        self._logger = logger

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        try:
            return self.app(environ, start_response)
        except Exception as error:
            if self._logger is not None:
                self._logger.exception(error)
            response_body = self.handler.get_raw_error(error)
            response_headers = [
                ("Content-Type", self.handler.content_type),
                ("Content-Length", str(len(response_body))),
            ]
            start_response("200 OK", response_headers)
            return (response_body,)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        try:
            await self.app(scope, receive, send)
        except Exception as error:
            if self._logger is not None:
                self._logger.exception(error)
            ...  # todo



from __future__ import annotations

from logging import getLogger
from typing import Any
from typing import ClassVar
from typing import Type

from drakaina import rpc_registry
from drakaina.exceptions import InternalServerError
from drakaina.exceptions import RPCError
from drakaina.registries import RPC_SCHEMA
from drakaina.registries import RPCRegistry
from drakaina.serializers import BaseSerializer
from drakaina.serializers import JsonSerializer
from drakaina.types import OpenAPI
from drakaina.types import OpenRPC
from drakaina.utils import unwrap_func

__all__ = ("BaseRPCProtocol",)

log = getLogger()


class BaseRPCProtocol:
    """Base class for representing the remote procedure call (RPC) protocol.

    To implement your own RPC protocol, you must implement the `handle` method,
    which must accept two parameters - the incoming message of your protocol
    `rpc_request` and an environment object 'request' that contains information
    about the connection, the transport layer and other information not
    directly related to the rpc protocol.

    You must also specify a base error class and a default error class.
    If you implement your own error class hierarchy, you will need to map your
    implementation to drakaina error classes.

    :param registry:
        Registry of remote procedures.
        Default: `drakaina.registries.rpc_registry` (generic module instance)
    :param serializer:
        Serializer object.
    :param schema_serializer:
        The serializer object to serialize the schema.
        Default: `JsonSerializer` (stdlib.json)

    """

    __slots__ = (
        "registry",
        "serializer",
        "schema_serializer",
        "__schema",
    )

    base_error_class: ClassVar = RPCError
    """Base class for RPC protocol implementation."""

    default_error_class: ClassVar = InternalServerError
    """The default error class for representing internal exceptions."""

    # When you implement this class interface by implementing a child class,
    # you must map the user protocol exception classes to the generic
    # exception classes `drakaina.exceptions` in this class variable.
    _errors_map: ClassVar = {Exception: RPCError}

    def __init__(
        self,
        registry: RPCRegistry | None = None,
        serializer: BaseSerializer | None = None,
        schema_serializer: BaseSerializer | None = None,
    ):
        self.registry = registry if registry is not None else rpc_registry
        self.serializer = serializer
        self.schema_serializer = schema_serializer or JsonSerializer()

        self.__schema = None

    def handle_raw_request(
        self,
        raw_data: bytes,
        request: Any | None = None,
    ) -> bytes:
        """Accepts raw data, deserializes, processes the RPC request,
        and returns the serialized result.

        :param raw_data:
            Raw request data.
        :type raw_data:
        :param request:
            Request object or context data. Can be provided to
            a remote procedure.
        :type request: Any
        :return:
            Serialized RPC response data.

        """
        try:
            parsed_data = self.serializer.deserialize(raw_data)
        except Exception as exc:
            return self.get_raw_error(exc)

        response_data = self.handle(parsed_data, request=request)
        if response_data is None:
            return b""

        try:
            return self.serializer.serialize(response_data)
        except Exception as exc:
            return self.get_raw_error(exc)

    def get_raw_error(
        self,
        error: RPCError | Type[RPCError] | Exception | Type[Exception],
    ) -> bytes:
        """Returns the serialized error object.

        :param error:
            The instance or class of the error.
        :type error: RPCError | Type[RPCError] | Exception | Type[Exception]
        :return:
            Raw error data.

        """
        rpc_error = self.handle_error(error)
        return self.serializer.serialize(rpc_error.as_dict())

    def handle(self, rpc_request: Any, request: Any | None = None) -> Any:
        """Handles a procedure call.

        :param rpc_request:
            RPC request in protocol format.
        :param request:
            Optional parameter that can be passed as an
            argument to the procedure. By default, None will be passed.
        :return:
            Returns the result in protocol format.

        """
        raise NotImplementedError(
            "You must implement the `handle` method in the child class",
        )

    def handle_error(
        self,
        error: RPCError | Type[RPCError] | Exception | Type[Exception],
    ) -> RPCError:
        """Returns an exception object corresponding to the RPC protocol.

        :param error:
            The instance or class of the error.
        :type error: RPCError | Type[RPCError] | Exception | Type[Exception]
        :returns: Protocol specific error object.
        :rtype: RPCError

        """

        print("handle_error", type(error), f"{error=}")
        if isinstance(error, type) and issubclass(error, Exception):
            error = error()

        if isinstance(error, self.base_error_class):
            return error
        else:
            # Try to get mapped error class
            rpc_error_class = self._errors_map.get(type(error))
            # If nothing is retrieved, try to get the mapped error class
            #  from the base error classes
            if rpc_error_class is None and isinstance(error, RPCError):
                for base_error_class in type(error).__mro__:
                    if base_error_class in self._errors_map:
                        rpc_error_class = self._errors_map[base_error_class]
                        break
            if rpc_error_class is None:
                rpc_error_class = self.default_error_class

            try:
                error_message = error.message or ""
            except AttributeError:
                error_message = ""

            return rpc_error_class(error_message)

    @property
    def content_type(self) -> str:
        return self.serializer.content_type

    @property
    def schema_content_type(self) -> str:
        return self.schema_serializer.content_type

    def schema(self) -> dict:
        """Simple schema of service."""
        if self.__schema is None:
            self.__schema = {}
            for method_name, procedure_ in self.registry.items():
                procedure = unwrap_func(procedure_)
                if hasattr(procedure, RPC_SCHEMA):
                    self.__schema[method_name] = getattr(procedure, RPC_SCHEMA)
        return self.__schema

    def get_raw_schema(self) -> bytes:
        return self.schema_serializer.serialize(self.schema())

    def openrpc_schema(self) -> OpenRPC:
        """Implementation of OpenRPC specification.

        https://spec.open-rpc.org/

        """
        raise NotImplementedError()

    def get_raw_openrpc_schema(self) -> bytes:
        # return self.schema_serializer.serialize(self.openrpc_schema())
        try:
            return self.schema_serializer.serialize(self.openrpc_schema())
        except Exception as err:
            import traceback

            print("get_raw_openrpc_schema", str(err))
            print(traceback.format_exc())

            log.exception(err)
            raise err

    def openapi_schema(self) -> OpenAPI:
        """Implementation of OpenAPI specification.

        https://spec.openapis.org/oas/latest.html

        """
        raise NotImplementedError()

    def get_raw_openapi_schema(self) -> bytes:
        # return self.schema_serializer.serialize(self.openapi_schema())
        try:
            return self.schema_serializer.serialize(self.openapi_schema())
        except Exception as err:
            import traceback

            print("get_raw_openrpc_schema", str(err))
            print(traceback.format_exc())

            log.exception(err)
            raise err



from __future__ import annotations

from functools import partial
from typing import Callable
from typing import Iterable
from typing import Optional

from django.http import HttpRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseNotAllowed
from django.utils.module_loading import autodiscover_modules
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from drakaina.exceptions import BadRequestError
from drakaina.rpc_protocols import BaseRPCProtocol
from drakaina.rpc_protocols import JsonRPCv2


class RPCView(View):
    """Django class based view implements JSON-RPC"""

    http_method_names = ["post", "options"]
    handler: BaseRPCProtocol
    content_type: str

    @classmethod
    def as_view(
        cls,
        autodiscover: str | Iterable[str] = "rpc_methods",
        handler: Optional[BaseRPCProtocol] = None,
        allowed_methods: Optional[Iterable[str]] = None,
        **initkwargs,
    ) -> Callable:
        """

        :param autodiscover: The name(s) submodule(s) in a django applications
                             in which the RPC methods are defined.
        :type autodiscover: str
        :param handler: RPC protocol implementation.
        :type handler: BaseRPCProtocol
        :param allowed_methods: Allowed http methods. Default: POST, OPTIONS.
        :type allowed_methods:
        :param initkwargs:
        :return: RPCView instance.

        """
        cls.handler = handler if handler is not None else JsonRPCv2()
        cls.content_type = cls.handler.content_type
        if allowed_methods is not None:
            cls.http_method_names = allowed_methods
        view = super().as_view(**initkwargs)
        view.cls = cls

        # Scan specified sub-modules
        autodiscover_modules(autodiscover)

        return csrf_exempt(view)

    def post(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        response_headers = {}
        if request.content_type != self.content_type:
            response = partial(HttpResponse, status=415)
            content = self.handler.get_raw_error(BadRequestError())
        elif len(request.body) > int(request.headers["Content-Length"]):
            response = HttpResponseBadRequest
            content = self.handler.get_raw_error(BadRequestError())
        else:
            content = self.handler.handle_raw_request(
                request.body,
                request=request,
            )
            if hasattr(request, "response"):
                _response = request.response
                if isinstance(_response, dict) and "headers" in _response:
                    for header in _response["headers"]:
                        response_headers[header[0]] = header[1]
            response = partial(HttpResponse, headers=response_headers)

        return response(content=content, content_type=self.content_type)

    def http_method_not_allowed(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        return HttpResponseNotAllowed(
            permitted_methods=(m.upper() for m in self.http_method_names),
            content=self.handler.get_raw_error(BadRequestError()),
            content_type=self.content_type,
        )

    def options(self, request, *args, **kwargs):
        return HttpResponse(
            headers={
                "Allow": ", ".join((m.upper() for m in self.http_method_names)),
            },
        )



from __future__ import annotations

from asyncio import iscoroutinefunction
from functools import update_wrapper
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import TypeVar

from drakaina.exceptions import ForbiddenError
from drakaina.registries import RPC_REGISTRY
from drakaina.registries import RPCRegistry
from drakaina.types import AnyRequest
from drakaina.types import Comparator
from drakaina.utils import iterable_str_arg

__all__ = (
    "RPCRegistry",
    "rpc_registry",
    "remote_procedure",
    "check_permissions",
    "login_required",
    "match_all",
    "match_any",
    "ENV_APP",
    "ENV_IS_AUTHENTICATED",
    "ENV_USER",
    "ENV_USER_ID",
    "ENV_AUTH_PAYLOAD",
    "ENV_AUTH_SCOPES",
)

# General registry
rpc_registry = RPCRegistry()

# Environ (Request) context
ENV_APP = "drakaina.app"
ENV_IS_AUTHENTICATED = "auth.is_authenticated"
ENV_USER = "user"
ENV_USER_ID = "user_id"
ENV_AUTH_PAYLOAD = "auth.payload"
ENV_AUTH_SCOPES = "auth.scopes"

T = TypeVar("T", bound=Callable)


def remote_procedure(
    name: Optional[str] = None,
    registry: Optional[RPCRegistry] = None,
    provide_request: Optional[bool] = None,
    metadata: Optional[dict[str, Any]] = None,
    **meta_options,
) -> Callable:
    """Decorator allow wrap function and define it as remote procedure.

    :param name:
        Procedure name. Default as function name.
    :type name: str
    :param registry:
        Procedure registry custom object
    :type registry: RPCRegistry
    :param provide_request:
        Provide a request object or context data (from the transport layer).
        If `True`, then the request object or context can be supplied to
        the procedure as a `request` argument.
    :type provide_request: bool
    :param metadata:
        Metadata that can be processed by middleware.
    :type metadata: dict[str, Any]
    :param meta_options:
        Metadata that can be processed by middleware.

    """

    def __decorator(
        procedure: T,
        _registry: RPCRegistry = None,
        _name: str = None,
        _provide_request: bool = None,
        _metadata: dict | None = None,
    ) -> T:
        """Returns a registered procedure"""

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                if not _provide_request:
                    if len(args) == 0:
                        kwargs.pop("request")
                    else:
                        scope, *args = args
                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                if not _provide_request:
                    if len(args) == 0:
                        kwargs.pop("request")
                    else:
                        environ, *args = args
                return procedure(*args, **kwargs)

        # Need to update the wrapper before registering in the registry
        decorated_procedure = update_wrapper(wrapper, procedure)

        if _registry is None:
            _registry = rpc_registry
        if _name is None:
            _name = procedure.__name__
        _registry.register_procedure(
            decorated_procedure,
            name=_name,
            provide_request=_provide_request,
            metadata=_metadata,
        )

        return decorated_procedure

    if callable(name):
        return __decorator(
            procedure=name,
            _registry=registry,
            _name=None,
            _provide_request=provide_request,
            _metadata={**(metadata or {}), **meta_options},
        )
    elif not isinstance(name, (str, type(None))):
        raise TypeError(
            "Expected first argument to be an str, a callable, or None",
        )

    def decorator(procedure):
        assert callable(procedure)
        return __decorator(
            procedure=procedure,
            _registry=registry,
            _name=name,
            _provide_request=provide_request,
            _metadata={**(metadata or {}), **meta_options},
        )

    return decorator


# TODO: may be implement `login_required` decorators for
#  Registries/RPC Handlers/apps for many instances.
#  But, first need implement nested registers, protocols and applications


def login_required(*_args) -> Callable:
    """Requires login decorator.

    Gives access to the procedure only to authenticated users.

    """

    def __decorator(procedure: T) -> T:
        """Returns a registered procedure"""
        registry: RPCRegistry = getattr(procedure, RPC_REGISTRY, None)
        if registry is None:
            raise TypeError(
                "Incorrect usage of decorator. Please use "
                "the `drakaina.remote_procedure` decorator first.",
            )

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)
                validate_authentication(request)
                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)
                validate_authentication(request)
                return procedure(*args, **kwargs)

        wrapper = update_wrapper(wrapper, procedure)
        registry.replace(procedure, wrapper)
        return wrapper

    if len(_args) > 0:
        if callable(_args[0]):
            return __decorator(_args[0])
        else:
            raise TypeError("Expected first argument to be a callable")

    def decorator(procedure):
        if not callable(procedure):
            raise TypeError("Expected first argument to be a callable")
        return __decorator(procedure)

    return decorator


def match_any(required: Iterable[str], provided: Iterable[str]) -> bool:
    if not (bool(required) or bool(provided)):  # If both are empty
        return True
    return any((scope in provided for scope in required))


def match_all(required: Iterable[str], provided: Iterable[str]) -> bool:
    return set(required).issubset(set(provided))


def check_permissions(
    scopes: str | Iterable[str],
    comparator: Comparator = match_any,
) -> Callable:
    """Permission decorator.

    Gives access to the procedure only to authorized users.

    """
    if not (
        isinstance(scopes, str)
        or (
            isinstance(scopes, Iterable)
            and all((isinstance(scope, str) for scope in scopes))
        )
    ):
        raise TypeError(
            "The `scopes` argument must be a string or Iterable[string]",
        )
    if not callable(comparator):
        raise TypeError(
            "The `comparator` argument must be a function that implements "
            "the `types.Comparator` interface",
        )

    procedure_scopes = iterable_str_arg(scopes)

    def __decorator(procedure: T) -> T:
        registry: RPCRegistry = getattr(procedure, RPC_REGISTRY, None)
        if registry is None:
            raise TypeError(
                "Incorrect usage of decorator. Please use "
                "the `drakaina.remote_procedure` decorator first.",
            )

        if iscoroutinefunction(procedure):
            # TODO: reference for async code `starlette.authentication.requires`

            async def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)  # noqa

                validate_authentication(request)
                validate_permissions(request, procedure_scopes, comparator)

                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                request = _get_request(*args, **kwargs)  # noqa

                validate_authentication(request)
                validate_permissions(request, procedure_scopes, comparator)

                return procedure(*args, **kwargs)

        wrapper = update_wrapper(wrapper, procedure)
        registry.replace(procedure, wrapper)
        return wrapper

    return __decorator


def _get_request(*args, **kwargs) -> Optional[AnyRequest]:
    if len(args) == 0:
        return kwargs.get("request")
    else:
        return args[0]


def _is_authenticated(request: AnyRequest) -> bool:
    if isinstance(request, (dict, Mapping)):
        return request.get(ENV_IS_AUTHENTICATED, False)
    else:
        return getattr(request, ENV_IS_AUTHENTICATED, False)


def validate_authentication(request: AnyRequest):
    """Validate authentication.

    :param request:
    :raise ForbiddenError:
    """
    if not _is_authenticated(request):
        raise ForbiddenError("Authentication required")


def _get_scopes(request: AnyRequest) -> Iterable[str]:
    if isinstance(request, (dict, Mapping)) or hasattr(request, "get"):
        scopes = request.get(ENV_AUTH_SCOPES, None)
    else:
        scopes = getattr(request, ENV_AUTH_SCOPES, None)
    if isinstance(scopes, (str, Iterable)):
        return iterable_str_arg(scopes)
    return ()


def validate_permissions(
    request: AnyRequest,
    procedure_scopes: Iterable[str],
    comparator: Comparator = match_any,
):
    """Validate permissions.

    :param request:
    :param procedure_scopes:
    :param comparator:
    :raise ForbiddenError:
    """
    user_scopes = _get_scopes(request)
    if not comparator(procedure_scopes, user_scopes):
        raise ForbiddenError("Forbidden")



from __future__ import annotations

from drakaina.types import ASGIApplication
from drakaina.types import ASGIReceive
from drakaina.types import ASGIScope
from drakaina.types import ASGISend
from drakaina.types import WSGIApplication
from drakaina.types import WSGIEnvironment
from drakaina.types import WSGIResponse
from drakaina.types import WSGIStartResponse


__all__ = ("BaseMiddleware",)


class BaseMiddleware:
    """Base class for middleware.

    It is a simple middleware for WSGI and ASGI, where the protocol
    will be chosen depending on the `is_async` parameter.

    :param app:
    :type app: WSGIApplication | ASGIApplication
    """

    __slots__ = ("app", "is_async", "__call__")

    def __init__(
        self,
        app: ASGIApplication | WSGIApplication,
        is_async: bool = False,
    ):
        self.app = app
        self.is_async = is_async
        if is_async:
            self.__call__ = self.__asgi_call__
        else:
            self.__call__ = self.__wsgi_call__

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        return self.app(environ, start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        await self.app(scope, receive, send)



from __future__ import annotations

from collections import OrderedDict
from collections.abc import Callable
from inspect import cleandoc
from inspect import Parameter
from inspect import Signature
from inspect import signature
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import KeysView
from typing import Mapping
from typing import ValuesView

from drakaina.types import MethodSchema
from drakaina.types import ParameterSchema
from drakaina.utils import short_description_by_name
from drakaina.utils import unwrap_func

try:
    from docstring_parser import parse as parse_doc
except ImportError:
    parse_doc = None

__all__ = (
    "is_rpc_procedure",
    "RPCRegistry",
    "RPC_REGISTRY",
    "RPC_NAME",
    "RPC_REGISTERED",
    "RPC_PROVIDE_REQUEST",
    "RPC_SCHEMA",
    "RPC_META",
)

# Reserved procedure argument names
RESERVED_KWARGS = ("cls", "self", "request")

# RPC procedure fields
RPC_REGISTRY = "__rpc_registry"
RPC_NAME = "__rpc_name"
RPC_REGISTERED = "__rpc_procedure"
RPC_PROVIDE_REQUEST = "__rpc_provide_request"
RPC_SCHEMA = "__rpc_schema"
RPC_META = "__rpc_metadata"


def is_rpc_procedure(func: Callable) -> bool:
    return hasattr(unwrap_func(func), RPC_REGISTERED)


class RPCRegistry(Mapping[str, Callable[..., Any]]):
    """Registry of remote procedures."""

    _remote_procedures: dict[str, Callable[..., Any]]

    def __init__(self):
        self._remote_procedures = {}

    def register_procedure(
        self,
        procedure: Callable[..., Any],
        name: str | None = None,
        provide_request: bool | None = None,
        metadata: dict | None = None,
    ):
        """Register a function as a remote procedure.

        :param procedure: Registered procedure.
        :param name: Procedure name. Default as function name.
        :param provide_request:
            If `True`, then the request object or context can be supplied to
            the procedure as a `request` argument.
        :param metadata:
            Metadata that can be processed by middleware or an extension.

        """
        assert callable(procedure)

        procedure_name = procedure.__name__ if name is None else name
        if procedure_name.startswith("rpc."):
            raise ValueError(
                "Method names that begin with 'rpc.' are reserved for "
                "rpc internal methods and extensions and MUST NOT be used "
                "for anything else.",
            )

        # Inspection of the procedure interface and its documentation.
        schema = self.inspect_procedure(procedure, metadata)
        schema["name"] = procedure_name

        self.__register_callable(
            func=procedure,
            name=procedure_name,
            provide_request=provide_request,
            schema=schema,
            metadata=metadata,
        )

    def register_rpc_extension(
        self,
        extension_procedure: Callable[..., Any],
        extension_name: str,
        provide_request: bool | None = None,
        metadata: dict | None = None,
    ):
        """Register a function as an RPC extension.

        :param extension_procedure: Registered RPC extension.
        :param extension_name: Procedure name.
        :param provide_request:
            If `True`, then the request object or context can be supplied to
            the procedure as a `request` argument.
        :param metadata:
            Metadata that can be processed by middleware or an extension.
        """
        assert callable(extension_procedure)
        assert extension_name and extension_name.startswith("rpc.")

        # TODO: need inspect schema ?

        self.__register_callable(
            func=extension_procedure,
            name=extension_name,
            provide_request=provide_request,
            metadata=metadata,
        )

    @staticmethod
    def inspect_procedure(
        procedure: Callable[..., Any],
        metadata: dict = None,
    ) -> MethodSchema:
        """Inspects the signature and generates a schematic of the function.

        :param procedure:
            Procedure for inspection.
        :param metadata:
            The metadata specified when registering the procedure.
            May contain additional information about the signature or
            documentation about the procedure.
        :return:
            Returns the schema for the given procedure filled with
            the found data.

        """
        if "method_schema" in metadata:
            schema = metadata["method_schema"]
        else:
            schema = MethodSchema(parameters=OrderedDict(), result={})

        if callable(parse_doc):
            docstring = parse_doc(getattr(procedure, "__doc__", ""))
            docstring.params_dict = {p.arg_name: p for p in docstring.params}
            if docstring.short_description:
                schema.setdefault(
                    "short_description",
                    docstring.short_description,
                )
            if docstring.long_description:
                schema.setdefault("description", docstring.long_description)
        else:
            docstring = None
            schema.setdefault(
                "short_description",
                short_description_by_name(procedure.__name__),
            )
            schema.setdefault(
                "description",
                metadata.get("description")
                or cleandoc(getattr(procedure, "__doc__", None) or ""),
            )

        proc_signature = signature(procedure)
        # Get parameters info from signature
        for parameter in proc_signature.parameters.values():
            if parameter.name in RESERVED_KWARGS:
                # Skip all reserved args
                continue

            if parameter.name in schema["parameters"]:
                param_schema = schema["parameters"][parameter.name]
            else:
                param_schema = ParameterSchema(
                    name=parameter.name,
                    kind=parameter.kind,  # noqa
                )

            # Set type
            if parameter.annotation is not Parameter.empty:
                if isinstance(parameter.annotation, type):
                    _type = parameter.annotation.__name__
                else:
                    _type = parameter.annotation
                param_schema.setdefault("type", _type)

            # Set default value
            if parameter.default is not Parameter.empty:
                param_schema.setdefault("default", parameter.default)
                param_schema.setdefault("required", False)
            else:
                param_schema.setdefault("required", True)

            # Get values from docstring
            if docstring and parameter.name in docstring.params_dict:
                param_doc = docstring.params_dict[parameter.name]
                param_schema.setdefault("type", param_doc.type_name)
                param_schema.setdefault("description", param_doc.description)

            # param_schema.setdefault(
            #     "description",
            #     short_description_by_name(parameter.name),
            # )

            # Add in method schema
            schema["parameters"].setdefault(parameter.name, param_schema)

        # Get return info from signature
        if proc_signature.return_annotation is not Signature.empty:
            # Set return type
            if isinstance(proc_signature.return_annotation, type):
                _type = proc_signature.return_annotation.__name__
            else:
                _type = proc_signature.return_annotation
            schema["result"].setdefault("type", _type)

        # Get values from docstring
        if docstring:
            # Return info
            returns = docstring.returns
            if returns and returns.return_name:
                schema["result"].setdefault("name", returns.return_name)
            if returns and returns.type_name:
                schema["result"].setdefault("type", returns.type_name)
            if returns and returns.description:
                schema["result"].setdefault("description", returns.description)
            # Exceptions info
            raises = docstring.raises
            if raises:
                schema["errors"] = []
                for exc in raises:
                    exc_schema = {"type": exc.type_name}
                    if exc.description:
                        exc_schema["description"] = exc.description
                    schema["errors"].append(exc_schema)
            # Deprecation info
            if docstring.deprecation:  # TODO
                # {
                #     "version": docstring.deprecation.version,
                #     "description": docstring.deprecation.description,
                # }
                ...
            if docstring.examples:  # TODO
                # {
                #     "snippet": docstring.deprecation.snippet,
                #     "description": docstring.deprecation.description,
                # }
                ...

        return schema

    def __register_callable(
        self,
        func: Callable[..., Any],
        name: str,
        provide_request: bool = False,
        schema: MethodSchema | dict | None = None,
        metadata: dict | None = None,
    ):
        """Placing the function in the registry with the specified name.

        :param func: Placed function.
        :param name: Name of function.
        :param provide_request:
            If `True`, then the request object or context can be supplied to
            the procedure as a `request` argument.
        :param schema: The function schema needed to create the documentation.
        :param metadata:
            Metadata that can be processed by middleware or an extension.

        """
        if name in self._remote_procedures:
            raise ValueError(
                "A procedure with this name is already registered.",
            )

        # The loop is for applying attributes to wrapped functions.
        # If the provided function is not wrapped, the attributes will be
        # applied once
        _func = func
        while _func is not None:
            setattr(_func, RPC_REGISTRY, self)
            setattr(_func, RPC_REGISTERED, True)
            setattr(_func, RPC_NAME, name)
            setattr(_func, RPC_PROVIDE_REQUEST, provide_request)
            setattr(_func, RPC_SCHEMA, schema or {})
            setattr(_func, RPC_META, metadata or {})
            # If it's a wrapped function
            _func = getattr(_func, "__wrapped__", None)

        self._remote_procedures[name] = func

    def replace(
        self,
        procedure: Callable[..., Any],
        new_procedure: Callable[..., Any],
    ):
        """Special method for use in cases where
        a decorator wrap procedure is required.

        :param procedure:
        :param new_procedure:

        """
        procedure_name = getattr(procedure, RPC_NAME, None)
        if procedure_name is None:
            raise TypeError(f"Callable {procedure} is not an rpc procedure.")
        # TODO: need duplicate __rpc attributes?
        self._remote_procedures[procedure_name] = new_procedure

    def __getitem__(self, key: str) -> Callable[..., Any] | None:
        # Intentionally suppress the rise of the exception,
        # and simply return None if there is no specified procedure.
        return self._remote_procedures.get(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self._remote_procedures)

    def __len__(self) -> int:
        return len(self._remote_procedures)

    def __contains__(self, item: str) -> bool:
        if not isinstance(item, str):
            raise TypeError(
                f"The procedure name must be a str, not {type(item)}.",
            )
        return item in self._remote_procedures.__contains__()

    def keys(self) -> KeysView[str]:
        return self._remote_procedures.keys()

    def items(self) -> Iterable[tuple[str, Callable[..., Any]]]:
        return self._remote_procedures.items()

    def values(self) -> ValuesView[Callable[..., Any]]:
        return self._remote_procedures.values()

    def get(
        self,
        key: str,
        default: Callable[..., Any] | None = None,
    ) -> Callable[..., Any] | None:
        if not isinstance(key, str):
            raise TypeError(
                f"The procedure name must be a str, not {type(key)}.",
            )
        if default is not None and not callable(default):
            raise TypeError("The default argument must be a callable.")
        return self._remote_procedures.get(key, default)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, self.__class__)
            and self._remote_procedures == other._remote_procedures
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)



from __future__ import annotations

from collections import OrderedDict
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import MutableSequence
from collections.abc import Sequence
from sys import version_info
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Literal
from typing import Protocol
from typing import Tuple
from typing import Type
from typing import TypedDict
from typing import Union

if version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

if version_info < (3, 11):
    from enum import Enum
    from typing_extensions import NotRequired

    # Copied from python 3.11

    class ReprEnum(Enum):
        """
        Only changes the repr(), leaving str() and format()
        to the mixed-in type.
        """

    class IntEnum(int, ReprEnum):
        """
        Enum where members are also (and must be) ints
        """

    class StrEnum(str, ReprEnum):
        """
        Enum where members are also (and must be) strings
        """

        def __new__(cls, *values):
            "values must already be of type `str`"
            if len(values) > 3:
                raise TypeError("too many arguments for str(): %r" % (values,))
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):
                    raise TypeError("%r is not a string" % (values[0],))
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(values[1], str):
                    raise TypeError(
                        "encoding must be a string, not %r" % (values[1],),
                    )
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(values[2], str):
                    raise TypeError(
                        "errors must be a string, not %r" % (values[2]),
                    )
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        def _generate_next_value_(name, start, count, last_values):
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()

else:
    from enum import IntEnum
    from enum import StrEnum
    from typing import NotRequired


"""
JSON-RPC types definitions
"""

JSONSimpleTypes: TypeAlias = Union[str, int, float, bool, None]
JSONTypes: TypeAlias = Union[
    JSONSimpleTypes,
    Mapping[str, JSONSimpleTypes],
    Sequence[JSONSimpleTypes],
]


class JSONRPCRequestObject(TypedDict):
    jsonrpc: Literal["2.0"]
    method: str
    params: NotRequired[list[JSONTypes] | dict[str, JSONTypes]]
    id: NotRequired[str | int | None]


class JSONRPCErrorObject(TypedDict):
    code: int
    message: str
    data: NotRequired[JSONTypes]


class JSONRPCResponseObject(TypedDict):
    jsonrpc: Literal["2.0"]
    result: NotRequired[JSONTypes]
    error: NotRequired[JSONRPCErrorObject]
    id: str | int | None


JSONRPCBatchRequestObject: TypeAlias = Sequence[JSONRPCRequestObject]
JSONRPCRequest: TypeAlias = Union[
    JSONRPCRequestObject,
    JSONRPCBatchRequestObject,
]
JSONRPCBatchResponseObject: TypeAlias = Sequence[JSONRPCResponseObject]
JSONRPCResponse: TypeAlias = Union[
    JSONRPCResponseObject,
    JSONRPCBatchResponseObject,
]


"""
Schema types
"""

json_schema_type_mapping = {
    None: "null",
    type(None): "null",
    "None": "null",
    "NoneType": "null",
    bool: "boolean",
    "bool": "boolean",
    int: "integer",  # number?
    "int": "integer",
    float: "number",
    "float": "number",
    str: "string",
    "str": "string",
    list: "array",
    "list": "array",
    dict: "object",
    "dict": "object",
}


def get_json_schema_type(t: type | str) -> str:
    result = json_schema_type_mapping.get(t)
    if result is None:
        result = str(t)
    return result
    # if isinstance(t, type):
    #     result = json_schema_type_mapping.get(t)
    #     if result is None:
    #         result = str(t)
    #     return result


_JSONSchema = TypedDict(
    "_JSONSchema",
    {
        "$id": NotRequired[str],
        "$ref": NotRequired[str],
        "$dynamicRef": NotRequired[str],
        "$dynamicAnchor": NotRequired[str],
        "$schema": NotRequired[str],
    },
)


class JSONSchema(_JSONSchema):
    """JSON schema.

    https://datatracker.ietf.org/doc/html/draft-bhutton-json-schema-00

    """

    base: NotRequired[str]
    type: NotRequired[str | Sequence[str]]
    description: NotRequired[str]
    format: NotRequired[str]
    required: NotRequired[list[str]]
    default: NotRequired[str]
    properties: NotRequired[Mapping[str, JSONSchema]]


Reference = TypedDict(
    "Reference",
    {
        "$ref": str,  # noqa
        "summary": NotRequired[str],
        "description": NotRequired[str],
    },
)


class ParameterSchema(TypedDict):
    name: NotRequired[str]
    short_description: NotRequired[str]
    description: NotRequired[str]
    type: NotRequired[str]
    default: NotRequired[Any]
    required: NotRequired[bool]
    """Default: False"""
    deprecated: NotRequired[bool]
    """Default: False"""


class MethodSchema(TypedDict):
    name: NotRequired[str]
    short_description: NotRequired[str]
    description: NotRequired[str]
    parameters: NotRequired[OrderedDict[str, ParameterSchema]]
    result: NotRequired[ParameterSchema]
    notification: NotRequired[bool]
    deprecated: NotRequired[bool]
    errors: NotRequired[dict[str, str]]
    example: NotRequired[Any]
    examples: NotRequired[Sequence[Any]]
    # tags: NotRequired[list[ORPCTag | Reference]]
    # paramStructure: NotRequired[Literal["by-name", "by-position", "either"]]
    # TODO: **MethodSchema(errors, examples, tags, )


"""
Schema types - OpenRPC
https://spec.open-rpc.org/
"""

SUPPORTED_OPENRPC_VERSION = "1.3.1"


class OpenRPC(TypedDict):
    openrpc: Literal["1.3.1"]
    info: ORPCInfo
    methods: list[ORPCMethod | Reference]
    servers: NotRequired[list[ORPCServer]]
    components: NotRequired[ORPCComponents]
    externalDocs: NotRequired[ORPCExtDoc]


class ORPCInfo(TypedDict):
    title: str
    version: str
    description: NotRequired[str]
    termsOfService: NotRequired[str]
    contact: NotRequired[ORPCContact]
    license: NotRequired[ORPCLicense]


class ORPCContact(TypedDict):
    name: NotRequired[str]
    url: NotRequired[str]
    email: NotRequired[str]


class ORPCLicense(TypedDict):
    name: str
    url: NotRequired[str]


class ORPCServer(TypedDict):
    name: str
    url: str
    """Runtime Expression"""
    summary: NotRequired[str]
    description: NotRequired[str]
    variables: NotRequired[Mapping[str, ORPCServerVar]]


class ORPCServerVar(TypedDict):
    default: str
    enum: NotRequired[list[str]]
    description: NotRequired[str]


class ORPCMethod(TypedDict):
    name: str
    params: list[ORPCContentDescriptor | Reference]
    result: NotRequired[ORPCContentDescriptor | Reference]
    tags: NotRequired[list[ORPCTag | Reference]]
    summary: NotRequired[str]
    description: NotRequired[str]
    externalDocs: NotRequired[ORPCExtDoc]
    deprecated: NotRequired[bool]
    """Default: False"""
    servers: NotRequired[list[ORPCServer]]
    errors: NotRequired[list[JSONRPCErrorObject | Reference]]
    links: NotRequired[list[ORPCLink | Reference]]
    paramStructure: NotRequired[Literal["by-name", "by-position", "either"]]
    examples: NotRequired[list[ORPCExamplePairing]]


class ORPCContentDescriptor(TypedDict):
    name: str
    schema: JSONSchema
    summary: NotRequired[str]
    description: NotRequired[str]
    required: NotRequired[bool]
    """Default: False"""
    deprecated: NotRequired[bool]
    """Default: False"""


class ORPCExamplePairing(TypedDict):
    name: NotRequired[str]
    summary: NotRequired[str]
    description: NotRequired[str]
    params: NotRequired[list[ORPCExample | Reference]]
    result: NotRequired[ORPCExample | Reference]


class ORPCExample(TypedDict):
    name: NotRequired[str]
    summary: NotRequired[str]
    description: NotRequired[str]
    value: NotRequired[Any]
    externalValue: NotRequired[str]


class ORPCLink(TypedDict):
    name: str
    summary: NotRequired[str]
    description: NotRequired[str]
    method: NotRequired[str]
    params: NotRequired[Mapping[str, Any | str]]
    """Mapping[str, Any | 'Runtime Expression']"""
    server: NotRequired[ORPCServer]


class ORPCComponents(TypedDict):
    contentDescriptors: NotRequired[Mapping[str, ORPCContentDescriptor]]
    schemas: NotRequired[Mapping[str, JSONSchema]]
    examples: NotRequired[Mapping[str, ORPCExample]]
    links: NotRequired[Mapping[str, ORPCLink]]
    errors: NotRequired[Mapping[str, JSONRPCErrorObject]]
    examplePairingObjects: NotRequired[Mapping[str, ORPCExamplePairing]]
    tags: NotRequired[Mapping[str, ORPCTag]]


class ORPCTag(TypedDict):
    name: str
    summary: NotRequired[str]
    description: NotRequired[str]
    externalDocs: NotRequired[ORPCExtDoc]


class ORPCExtDoc(TypedDict):
    url: str
    description: NotRequired[str]


class ORPC:
    SUPPORTED_VERSION = SUPPORTED_OPENRPC_VERSION
    OpenRPC = OpenRPC
    Info = ORPCInfo
    Contact = ORPCContact
    License = ORPCLicense
    Server = ORPCServer
    ServerVar = ORPCServerVar
    Method = ORPCMethod
    ContentDescriptor = ORPCContentDescriptor
    Schema = JSONSchema
    ExamplePairing = ORPCExamplePairing
    Example = ORPCExample
    Link = ORPCLink
    Components = ORPCComponents
    Tag = ORPCTag
    ExtDoc = ORPCExtDoc
    Reference = Reference

    type_mapping = get_json_schema_type

    @classmethod
    def method_by_schema(
        cls,
        method_schema: MethodSchema,
        default_name: str | None = None,
    ) -> ORPCMethod:
        if default_name is None:
            default_name = method_schema["name"]
        method = ORPCMethod(
            name=default_name,
            params=[],
            result=ORPCContentDescriptor(),
            deprecated=method_schema.get("deprecated", False),
        )
        if "short_description" in method_schema:
            method["summary"] = method_schema["short_description"]
        if "description" in method_schema:
            method["description"] = method_schema["description"]

        # Fill parameters schema
        parameters = method_schema.get("parameters", {})
        for parameter_name, parameter in parameters.items():
            method["params"].append(
                cls.content_descriptor_by_parameter(
                    parameter,
                    default_name=parameter_name,
                ),
            )

        # Fill result schema
        if "result" in method_schema:
            method["result"] = cls.content_descriptor_by_result(
                method_schema["result"],
            )

        return method

    @staticmethod
    def content_descriptor_by_parameter(
        parameter: ParameterSchema,
        default_name: str | None = None,
    ) -> ORPCContentDescriptor:
        cd = ORPCContentDescriptor(
            name=parameter.get("name", default_name),
            schema=JSONSchema(),
            required=parameter.get("required", False),
            deprecated=parameter.get("deprecated", False),
        )
        if "type" in parameter:
            cd["schema"]["type"] = get_json_schema_type(
                parameter["type"],
            )
        if "default" in parameter:
            cd["schema"]["default"] = parameter["default"]
        if "short_description" in parameter:
            cd["summary"] = parameter["short_description"]
        if "description" in parameter:
            cd["description"] = parameter["description"]

        return cd

    @staticmethod
    def content_descriptor_by_result(
        result_schema: dict,
        default_name: str = "result",
    ) -> ORPCContentDescriptor:
        cd = ORPC.ContentDescriptor(
            name=result_schema.get("name", default_name),
            schema=ORPC.Schema(),
        )
        if "type" in result_schema:
            cd["schema"]["type"] = get_json_schema_type(
                result_schema["type"],
            )
        if "description" in result_schema:
            cd["summary"] = result_schema["description"]
            cd["description"] = result_schema["description"]

        return cd


"""
Schema types - OpenAPI
https://github.com/OAI/OpenAPI-Specification
"""

SUPPORTED_OPENAPI_VERSION = "3.0.0"


class OpenAPI(TypedDict):
    openapi: Literal["3.1.0"]
    info: OASInfo
    jsonSchemaDialect: NotRequired[str]
    """URI"""
    servers: NotRequired[list[OASServer]]
    paths: NotRequired[OASPaths]
    """OASPaths["/{path}", OASPathItem]"""
    webhooks: NotRequired[Mapping[str, OASPathItem | Reference]]
    components: NotRequired[OASComponents]
    security: NotRequired[list[OASSecurityRequirement]]
    tags: NotRequired[list[OASTag]]
    externalDocs: NotRequired[OASExtDoc]


class OASInfo(TypedDict):
    title: str
    version: str
    summary: NotRequired[str]
    description: NotRequired[str]
    termsOfService: NotRequired[str]
    contact: NotRequired[OASContact]
    license: NotRequired[OASLicense]


class OASContact(TypedDict):
    name: NotRequired[str]
    url: NotRequired[str]
    email: NotRequired[str]


class OASLicense(TypedDict):
    name: str
    identifier: NotRequired[str]
    url: NotRequired[str]


class OASServer(TypedDict):
    url: str
    description: NotRequired[str]
    variables: NotRequired[Mapping[str, OASServerVar]]


class OASServerVar(TypedDict):
    default: str
    enum: NotRequired[list[str]]
    description: NotRequired[str]


class OASComponents(TypedDict):
    schemas: NotRequired[Mapping[str, OASSchema]]
    responses: NotRequired[Mapping[str, OASResponse | Reference]]
    parameters: NotRequired[Mapping[str, OASParameter | Reference]]
    examples: NotRequired[Mapping[str, OASExample | Reference]]
    requestBodies: NotRequired[Mapping[str, OASRequestBody | Reference]]
    headers: NotRequired[Mapping[str, OASHeader | Reference]]
    securitySchemes: NotRequired[Mapping[str, OASSecurityScheme | Reference]]
    links: NotRequired[Mapping[str, OASLink | Reference]]
    callbacks: NotRequired[Mapping[str, OASCallback | Reference]]
    pathItems: NotRequired[Mapping[str, OASPathItem | Reference]]


OASPathItemRef = TypedDict(
    "OASPathItemRef",
    {"$ref": NotRequired[str]},
)


class OASPathItem(OASPathItemRef):
    summary: NotRequired[str]
    description: NotRequired[str]
    get: NotRequired[OASOperation]
    put: NotRequired[OASOperation]
    post: NotRequired[OASOperation]
    delete: NotRequired[OASOperation]
    options: NotRequired[OASOperation]
    head: NotRequired[OASOperation]
    patch: NotRequired[OASOperation]
    trace: NotRequired[OASOperation]
    servers: NotRequired[OASServer]
    parameters: NotRequired[OASParameter | Reference]


OASPaths = TypedDict("OASPaths", {"/{path}": OASPathItem})


class OASOperation(TypedDict):
    tags: NotRequired[list[str]]
    summary: NotRequired[str]
    description: NotRequired[str]
    externalDocs: NotRequired[OASExtDoc]
    operationId: NotRequired[str]
    parameters: NotRequired[Sequence[OASParameter | Reference]]
    requestBody: NotRequired[OASRequestBody | Reference]
    responses: NotRequired[OASResponses]
    callbacks: NotRequired[Mapping[str, OASCallback | Reference]]
    deprecated: NotRequired[bool]
    security: NotRequired[Sequence[OASSecurityRequirement]]
    servers: NotRequired[Sequence[OASServer]]


class OASExtDoc(TypedDict):
    url: str
    description: NotRequired[str]


class OASParameterLocation(StrEnum):
    PATH = "path"
    QUERY = "query"
    HEADER = "header"
    COOKIE = "cookie"


OASParameterIn = TypedDict(
    "OASParameterIn",
    {"in": OASParameterLocation},
)


class OASParameter(OASParameterIn):
    name: str
    # in: str
    description: NotRequired[str]
    required: NotRequired[bool]
    """required = True if `in` == 'path'"""
    deprecated: NotRequired[bool]
    allowEmptyValue: NotRequired[bool]
    style: NotRequired[str]
    explode: NotRequired[bool]
    allowReserved: NotRequired[bool]
    schema: NotRequired[OASSchema]
    example: NotRequired[Any]
    examples: NotRequired[Mapping[str, OASExample | Reference]]
    content: NotRequired[Mapping[str, OASMediaType]]
    """A parameter MUST contain either a schema property, or
    a content property, but not both."""


class OASRequestBody(TypedDict):
    content: Mapping[str, OASMediaType]
    description: NotRequired[str]
    required: NotRequired[bool]


class OASMediaType(TypedDict):
    schema: NotRequired[OASSchema]
    example: NotRequired[Any]
    examples: NotRequired[Mapping[str, OASExample | Reference]]
    encoding: NotRequired[Mapping[str, OASEncoding]]


class OASEncoding(TypedDict):
    contentType: NotRequired[str]
    headers: NotRequired[Mapping[str, OASHeader | Reference]]
    style: NotRequired[str]
    explode: NotRequired[bool]
    allowReserved: NotRequired[bool]


class OASResponse(TypedDict):
    description: str
    headers: NotRequired[Mapping[str, OASHeader | Reference]]
    content: NotRequired[Mapping[str, OASMediaType]]
    links: NotRequired[Mapping[str, OASLink | Reference]]


OASResponses = TypedDict(
    "OASResponses",
    {
        "{status_code}": NotRequired[Union[OASResponse, Reference]],
        "default": NotRequired[Union[OASResponse, Reference]],
    },
)


OASCallback = Dict[str, Union[OASPathItem, Reference]]


class OASExample(TypedDict):
    summary: NotRequired[str]
    description: NotRequired[str]
    value: NotRequired[Any]
    externalValue: NotRequired[str]


class OASLink(TypedDict):
    operationRef: NotRequired[str]
    operationId: NotRequired[str]
    parameters: NotRequired[Mapping[str, Any | str]]
    """NotRequired[Mapping[str, Any | {expression}]]"""
    description: NotRequired[str]
    server: NotRequired[OASServer]


class OASHeader(TypedDict):
    description: NotRequired[str]
    required: NotRequired[bool]
    deprecated: NotRequired[bool]
    allowEmptyValue: NotRequired[bool]
    style: NotRequired[str]
    explode: NotRequired[bool]
    allowReserved: NotRequired[bool]
    schema: NotRequired[OASSchema]
    example: NotRequired[Any]
    examples: NotRequired[Mapping[str, OASExample | Reference]]
    content: NotRequired[Mapping[str, OASMediaType]]
    """A parameter MUST contain either a schema property, or
    a content property, but not both."""


class OASTag(TypedDict):
    name: str
    description: NotRequired[str]
    externalDocs: NotRequired[OASExtDoc]


class OASSchema(JSONSchema):
    discriminator: NotRequired[OASDiscriminator]
    xml: NotRequired[OASXML]
    externalDocs: NotRequired[OASExtDoc]
    example: NotRequired[Any]


class OASDiscriminator(TypedDict):
    propertyName: str
    mapping: NotRequired[Mapping[str, str]]


class OASXML(TypedDict):
    name: NotRequired[str]
    namespace: NotRequired[str]
    prefix: NotRequired[str]
    attribute: NotRequired[bool]
    wrapped: NotRequired[bool]


class OASSecuritySchemeType(StrEnum):
    API_KEY = "apiKey"


class OASSecuritySchemeLocation(StrEnum):
    QUERY = "query"
    HEADER = "header"
    COOKIE = "cookie"


OASSecuritySchemeIn = TypedDict(
    "OASSecuritySchemeIn",
    {"in": OASSecuritySchemeLocation},
)


class OASSecurityScheme(OASSecuritySchemeIn):
    type: OASSecuritySchemeType
    name: str
    scheme: str
    flows: OASOAuthFlows
    openIdConnectUrl: str
    description: NotRequired[str]
    bearerFormat: NotRequired[str]


class OASOAuthFlows(TypedDict):
    implicit: NotRequired[OASOAuthFlow]
    password: NotRequired[OASOAuthFlow]
    clientCredentials: NotRequired[OASOAuthFlow]
    authorizationCode: NotRequired[OASOAuthFlow]


class OASOAuthFlow(TypedDict):
    authorizationUrl: str
    tokenUrl: str
    scopes: str
    refreshUrl: NotRequired[str]


OASSecurityRequirement = Dict[str, list[str]]


class OAPI:
    SUPPORTED_VERSION = SUPPORTED_OPENAPI_VERSION
    OpenAPI = OpenAPI
    Info = OASInfo
    Contact = OASContact
    License = OASLicense
    Server = OASServer
    ServerVar = OASServerVar
    Components = OASComponents
    Paths = OASPaths
    PathItem = OASPathItem
    Operation = OASOperation
    ExtDoc = OASExtDoc
    ParameterLocation = OASParameterLocation
    Parameter = OASParameter
    RequestBody = OASRequestBody
    MediaType = OASMediaType
    Encoding = OASEncoding
    Responses = OASResponses
    Response = OASResponse
    Callback = OASCallback
    Example = OASExample
    Link = OASLink
    Header = OASHeader
    Tag = OASTag
    Reference = Reference
    Schema = OASSchema
    Discriminator = OASDiscriminator
    XML = OASXML
    SecurityScheme = OASSecurityScheme
    OAuthFlows = OASOAuthFlows
    OAuthFlow = OASOAuthFlow
    SecurityRequirement = OASSecurityRequirement

    type_mapping = get_json_schema_type

    """
    TODO: Need to refactor and improve the circuit generation
          with these methods.
    """

    @staticmethod
    def operation_by_schema(
        method_schema: MethodSchema,
        default_name: str | None = None,
    ) -> OASOperation:
        operation = OAPI.Operation(
            operationId=default_name or method_schema.get("name"),
            parameters=[],
            deprecated=method_schema.get("deprecated", False),
        )
        if "short_description" in method_schema:
            operation["summary"] = method_schema["short_description"]
        if "description" in method_schema:
            operation["description"] = method_schema["description"]

        # Fill request body
        operation["requestBody"] = OAPI.request_body_by_schema(
            method_schema,
        )

        # Fill result schema
        operation["responses"] = OAPI.responses_by_schema(method_schema)

        return operation

    @staticmethod
    def parameters_by_schema(method_schema: MethodSchema) -> list[OASParameter]:
        result = []
        parameters = method_schema.get("parameters", {})
        for param_name, parameter in parameters.items():
            param = OAPI.Parameter(
                name=parameter.get("name", param_name),
                # in=OAPI.ParameterLocation.?,
                schema=OAPI.Schema(),
                required=parameter.get("required", False),
                deprecated=parameter.get("deprecated", False),
            )
            if "type" in parameter:
                param["schema"]["type"] = OAPI.type_mapping(
                    parameter["type"],
                )
            if "default" in parameter:
                param["schema"]["default"] = parameter["default"]
            if "short_description" in parameter:
                param["summary"] = parameter["short_description"]
            if "description" in parameter:
                param["description"] = parameter["description"]

            result.append(param)

        return result

    @staticmethod
    def request_body_by_schema(method_schema: MethodSchema) -> OASRequestBody:
        parameters = method_schema.get("parameters", {}).items()
        media = OAPI.MediaType(
            schema=OAPI.Schema(
                type="object",
                properties={
                    "jsonrpc": {
                        "type": "string",
                        "description": "Protocol version",
                    },
                    "method": {
                        "type": "string",
                        "description": "Procedure (method) name",
                    },
                    "params": {
                        "type": "object",
                        "properties": {
                            param_name: {
                                # "type": OAPI.type_mapping(
                                #     parameter.get("type"),
                                # ),
                                "default": parameter.get("default", ""),
                                # "required": parameter.get(
                                #     "required",
                                #     False,
                                # ),
                                "description": parameter.get(
                                    "description",
                                    "",
                                ),
                                "deprecated": parameter.get(
                                    "deprecated",
                                    False,
                                ),
                            }
                            for param_name, parameter in parameters
                        },
                    },
                    "id": {
                        # "type": ["string", "integer"],  # TODO
                        "type": "string",
                    },
                },
                required=["jsonrpc", "method"],
                # x={
                #     "jsonrpc": "2.0",
                #     "method": method_name,
                #     "params": {},
                #     "id": 1,
                # },  # TODO <-
            ),
        )
        if "example" in method_schema:
            media["example"] = method_schema["example"]
        if "examples" in method_schema:
            # TODO: {str: OAPI.Example)
            media["examples"] = method_schema["examples"]
        return OASRequestBody(
            content={"application/json": media},
            description="JSON-RPC 2.0 Request Object",
            required=True,
        )

    @staticmethod
    def responses_by_schema(method_schema: MethodSchema) -> OASResponses:
        if "result" in method_schema:
            result_schema = method_schema["result"]
            jsonrpc_object = {
                "jsonrpc": OASSchema(type="string", default="2.0"),
                "result": OASSchema(),
                "id": OASSchema(type="string"),
            }
            if "type" in result_schema:
                jsonrpc_object["result"]["type"] = get_json_schema_type(
                    result_schema["type"],
                )

            media = OASMediaType(
                schema=OASSchema(type="object", properties=jsonrpc_object),
            )
            if "description" in result_schema:
                media["schema"]["description"] = result_schema["description"]
            # if "example" in result_schema:
            #     media["example"] = result_schema["example"]
            # if "examples" in result_schema:
            #     # TODO: {str: OAPI.Example)
            #     media["examples"] = result_schema["examples"]
        else:
            # If returns type is NoneType
            # TODO: Move to components
            media = OASMediaType(
                schema=OAPI.Schema(
                    type="object",
                    properties=dict(
                        jsonrpc=dict(type="string", default="2.0"),
                        result=dict(type="null"),
                        id=dict(type="string"),
                    ),
                ),
            )
        return OAPI.Responses(
            **{
                "200": OAPI.Response(
                    description="JSON-RPC 2.0 Response Object",
                    # TODO: content-type get from serializer
                    content={"application/json": media},
                    # TODO: required=True,
                    #  when method is not notification
                ),
            },
        )


"""
WSGI types definitions
PEP 3333 – Python Web Server Gateway Interface
https://peps.python.org/pep-3333/
"""

WSGIEnvironmentKeys: TypeAlias = Literal[
    # for CGI
    # https://datatracker.ietf.org/doc/html/draft-coar-cgi-v11-03
    "AUTH_TYPE",
    "CONTENT_LENGTH",
    "CONTENT_TYPE",
    "GATEWAY_INTERFACE",
    "PATH_INFO",
    "PATH_TRANSLATED",
    "QUERY_STRING",
    "REMOTE_ADDR",
    "REMOTE_HOST",
    "REMOTE_IDENT",
    "REMOTE_USER",
    "REQUEST_METHOD",
    "SCRIPT_NAME",
    "SERVER_NAME",
    "SERVER_PORT",
    "SERVER_PROTOCOL",
    "SERVER_SOFTWARE",
    # for WSGI
    "wsgi.errors",
    "wsgi.input",
    "wsgi.multiprocess",
    "wsgi.multithread",
    "wsgi.run_once",
    "wsgi.url_scheme",
    "wsgi.version",
]
# for framework needs
WSGIDrakainaKeys: TypeAlias = Literal[
    "drakaina.app",
    "drakaina.is_authenticated",
]
WSGIEnvironment: TypeAlias = MutableMapping[str, Any]
WSGIExceptionInfo: TypeAlias = Tuple[Type[BaseException], BaseException, Any]


class WSGIStartResponse(Protocol):
    def __call__(
        self,
        status: str,
        headers: MutableSequence[tuple[str, str]],
        exc_info: WSGIExceptionInfo | None = ...,
    ) -> Callable[[bytes], Any]:
        ...


WSGIResponse: TypeAlias = Iterable[bytes]
WSGIApplication: TypeAlias = Callable[
    [WSGIEnvironment, WSGIStartResponse],
    WSGIResponse,
]


class WSGIInputStream(Protocol):
    def read(self, size: int | None = None) -> bytes:
        ...

    def readline(self) -> bytes:
        ...

    def readlines(self, hint: Any | None) -> Iterable[bytes]:
        ...

    def __iter__(self) -> bytes:
        ...


class WSGIErrorsStream(Protocol):
    def flush(self) -> None:
        ...

    def write(self, s: str) -> None:
        ...

    def writelines(self, seq: Sequence[str]) -> None:
        ...


"""
ASGI types definitions
https://asgi.readthedocs.io/en/latest/
"""

ASGIScope: TypeAlias = MutableMapping[str, Any]
ASGIMessage: TypeAlias = MutableMapping[str, Any]

ASGIReceive: TypeAlias = Callable[[], Awaitable[ASGIMessage]]
ASGISend: TypeAlias = Callable[[ASGIMessage], Awaitable[None]]

ASGIApplication: TypeAlias = Callable[
    [ASGIScope, ASGIReceive, ASGISend],
    Awaitable[None],
]


"""
RSGI types definitions
https://github.com/emmett-framework/granian/blob/master/docs/spec/RSGI.md
"""


class RSGIScope:
    proto: Literal["http", "ws"] = "http"
    rsgi_version: Literal["1", "1.1", "2"]
    http_version: str
    server: str
    client: str
    scheme: Literal["http", "https"]
    method: str
    path: str
    query_string: str
    headers: Mapping[str, str]


class RSGIHTTPProtocol(Protocol):
    async def __call__(self, body: bytes) -> None:
        ...

    def response_empty(
        self,
        status: int,
        headers: list[tuple[str, str]],
        body: None,
    ) -> None:
        ...

    def response_str(
        self,
        status: int,
        headers: list[tuple[str, str]],
        body: str,
    ) -> None:
        ...

    def response_bytes(
        self,
        status: int,
        headers: list[tuple[str, str]],
        body: bytes,
    ) -> None:
        ...

    def response_file(
        self,
        status: int,
        headers: list[tuple[str, str]],
        body: str,
    ) -> None:
        ...


class RSGITransport(Protocol):
    async def receive(self) -> bytes:
        ...

    async def send_bytes(self, message: bytes) -> None:
        ...

    async def send_str(self, message: str) -> None:
        ...


class MessageKind(IntEnum):
    CLOSED = 0
    """Websocket closed by client"""
    BYTES = 1
    """Bytes message"""
    STRING = 2
    """String message"""


class RSGIWebsocketMessage:
    kind: Literal[0, 1, 2]
    data: bytes | str | None


class RSGIWSProtocol(Protocol):
    async def accept(self) -> RSGIWebsocketMessage:
        ...

    def close(self) -> None:
        ...


RSGIProtocol: TypeAlias = Union[RSGIHTTPProtocol, RSGIWSProtocol]


RSGIApplication: TypeAlias = Callable[
    [RSGIScope, RSGIProtocol],
    Awaitable[None],
]


"""
Helpful types
"""


class Comparator(Protocol):
    def __call__(
        self,
        required: Iterable[str],
        provided: Iterable[str],
    ) -> bool:
        ...


class ProxyRequest(MutableMapping):
    """A wrapper class for environment mapping.

    :param environment:

    TODO: check requestProxy on get_set_ENV_attrs
    """

    __slots__ = ("__environment",)

    def __init__(self, environment: ASGIScope | WSGIEnvironment):
        self.__environment = environment

    def __getitem__(self, item):
        return self.__environment[item]

    def __setitem__(self, key, value):
        self.__environment[key] = value

    def __delitem__(self, key):
        del self.__environment[key]

    def __iter__(self):
        return iter(self.__environment.keys())

    def __contains__(self, item):
        return item in self.__environment.keys()

    def __len__(self):
        return len(self.__environment)

    def keys(self):
        return self.__environment.keys()

    def values(self):
        return self.__environment.values()

    def items(self):
        return self.__environment.items()

    def get(self, key, default=None):
        return self.__environment.get(key, default)

    def clear(self):
        self.__environment.clear()

    def setdefault(self, key, default=None):
        self.__environment.setdefault(key, default)

    def pop(self, key, default=None):
        return self.__environment.pop(key, default)

    def popitem(self):
        return self.__environment.popitem()

    def copy(self):
        return self.__class__(self.__environment.copy())

    def update(self, *args, **kwargs):
        return self.__environment.update(*args, **kwargs)

    def __getattr__(self, item):
        if item in ("__environment", "_ProxyRequest__environment"):
            super().__getattribute__(item)
        return self.__environment[item]

    def __setattr__(self, key, value):
        if key in ("__environment", "_ProxyRequest__environment"):
            super().__setattr__(key, value)
        else:
            self.__environment[key] = value

    def __delattr__(self, item):
        del self.__environment[item]


AnyRequest: TypeAlias = Union[
    ASGIScope,
    RSGIScope,
    WSGIEnvironment,
    ProxyRequest,
]



""" Utility functions for package."""
from __future__ import annotations

import inspect
import re
from asyncio import iscoroutine
from decimal import Decimal
from datetime import date
from datetime import datetime
from datetime import timedelta
from functools import partial
from inspect import cleandoc
from json import JSONEncoder
from typing import Callable
from typing import Iterable
from typing import Literal


def iterable_str_arg(s: str | Iterable[str]) -> tuple[str]:
    if isinstance(s, str):
        s = s.replace(",", " ").strip().replace("  ", " ").split()
    return tuple(map(str, s))


def unwrap_func(func: Callable) -> Callable:
    _func = func
    while hasattr(_func, "__wrapped__") or isinstance(_func, partial):
        _func = getattr(_func, "__wrapped__", None) or _func.func
    return _func


def parse_doc_string(s: str) -> dict:
    parsed_doc = {"__doc__": ""}
    doc = cleandoc(s)

    ...

    return parsed_doc


def short_description_by_name(s: str) -> str:
    if isinstance(s, str):
        return re.sub(r"[\.-_]", " ", s).capitalize()


async def invoke(func):
    result = func()
    if iscoroutine(result):
        return await result
    return result


def get_cookies(cookie_str: str) -> dict[str, str]:
    if cookie_str is None:
        return {}

    # it may be a many matches?
    # wsgi_combined_cookie = ";".join(headers.getlist("HTTP_COOKIE"))

    _cookies = {}
    for cookie_item in cookie_str.split(";"):
        if "=" not in cookie_item or "DELETED" in cookie_item:
            continue
        name, value = cookie_item.strip().split("=", 1)
        _cookies[name] = value

    return _cookies


def set_cookie(
    name: str,
    value: str,
    expires: datetime | int | str | None = None,
    max_age: timedelta | int | None = None,
    domain: str | None = None,
    path: str | None = None,
    secure: bool | None = None,
    http_only: bool | None = None,
    same_site: Literal["Strict", "Lax", "None"] | None = None,
    flags: Iterable[str] | None = None,
) -> tuple[str, str]:
    """Returns tuple with cookie values prepared for ASGI/WSGI interface.

    RFC6265: https://datatracker.ietf.org/doc/html/rfc6265

    """
    if "__Secure" in name:
        secure = True
    if "__Host" in name:
        secure = True
        domain = None
        path = "/"
    value = value.replace(";", "")
    if expires:
        if isinstance(expires, int):
            expires = datetime.fromtimestamp(expires)
        if isinstance(expires, datetime):
            expires = expires.strftime("%a, %d %b %Y %H:%M:%S %Z")
        assert isinstance(expires, str)
        value += f"; Expires={expires}"
    if max_age:
        if isinstance(max_age, timedelta):
            max_age = max_age.days * 24 * 3600 + max_age.seconds
        assert isinstance(max_age, int) and max_age > 0
        value += f"; Max-Age={max_age}"
    if isinstance(domain, str):
        value += f"; Domain={domain}"
    if isinstance(path, str):
        value += f"; Path={path}"
    if secure:
        value += "; Secure"
    if http_only:
        value += "; HttpOnly"
    if same_site:
        assert same_site in ("Strict", "Lax", "None")
        if same_site == "None":
            assert secure is True
        value += f"; SameSite={same_site}"
    if flags:
        value += "; " + "; ".join(flags)

    return "Set-Cookie", f"{name}={value}".strip()


def mark_cookie_as_deleted(name: str) -> tuple[str, str]:
    expiry_string = "Thu, 01 Jan 1970 00:00:00 GMT"
    value = "DELETED"
    return "Set-Cookie", f"{name}={value}; Expires={expiry_string}"


def as_headers(cookie_dict: dict[str, str]):
    """start_response("200 OK", self.headers)"""
    _headers = []
    for k, v in cookie_dict.items():
        _headers.append(("Set-Cookie", f"{k}={v}"))
    return _headers


def match_path(item_path: str | re.Pattern[str], path_info: str) -> bool:
    """Check if the `path_info` (CGI) is equal to the given `item_path`
    of the element being checked.

    :param item_path: URL or regex pattern to match against the request path.
    :param path_info: Request path string.
    """

    if item_path is None:
        return True
    if isinstance(item_path, str):
        # or path_info.startswith(item_path) ?
        return path_info == item_path
    if isinstance(item_path, re.Pattern):
        return bool(item_path.match(path_info))

    return False


def check_request(path, entries):
    for pattern in entries:
        if re.match(pattern, path):
            return True

    return False


class DatetimeDecimalEncoder(JSONEncoder):
    """Encoder for datetime and decimal serialization.

    Usage: json.dumps(object, cls=DatetimeDecimalEncoder)
    NOTE: _iterencode does not work

    """

    def default(self, obj) -> str:
        """Encode JSON.

        :return: A JSON encoded string

        """
        if isinstance(obj, Decimal):
            return float(obj)

        if isinstance(obj, (datetime, date)):
            return obj.isoformat()

        return super().default(self, obj)


# validating parameters
def is_invalid_params(func, *args, **kwargs):
    """
    Method:
        Validate pre-defined criteria, if any is True - function is invalid
        0. func should be callable
        1. kwargs should not have unexpected keywords
        2. remove kwargs.keys from func.parameters
        3. number of args should be <= remaining func.parameters
        4. number of args should be >= remaining func.parameters less default
    """
    # For builtin functions inspect.signature(func) return error. If builtin
    # function generates TypeError, it is because of wrong parameters.
    if not inspect.isfunction(func):
        return True

    signature = inspect.signature(func)
    parameters = signature.parameters

    unexpected = set(kwargs.keys()) - set(parameters.keys())
    if len(unexpected) > 0:
        return True

    params = [
        parameter
        for name, parameter in parameters.items()
        if name not in kwargs
    ]
    params_required = [
        param for param in params if param.default is param.empty
    ]

    return not (len(params_required) <= len(args) <= len(params))



from __future__ import annotations

from copy import deepcopy
from functools import partial
from inspect import getmembers
from logging import getLogger
from typing import Any
from typing import ClassVar

from drakaina.exceptions import AuthenticationFailedError
from drakaina.exceptions import BadRequestError
from drakaina.exceptions import DeserializationError
from drakaina.exceptions import ForbiddenError
from drakaina.exceptions import InternalServerError
from drakaina.exceptions import InvalidParametersError
from drakaina.exceptions import InvalidPermissionsError
from drakaina.exceptions import InvalidTokenError
from drakaina.exceptions import NotFoundError
from drakaina.exceptions import RPCError
from drakaina.exceptions import SerializationError
from drakaina.registries import RPC_META
from drakaina.registries import RPC_SCHEMA
from drakaina.registries import RPCRegistry
from drakaina.rpc_protocols.base import BaseRPCProtocol
from drakaina.serializers import BaseSerializer
from drakaina.serializers import JsonSerializer
from drakaina.types import JSONRPCRequest
from drakaina.types import JSONRPCRequestObject
from drakaina.types import JSONRPCResponse
from drakaina.types import JSONRPCResponseObject
from drakaina.types import MethodSchema
from drakaina.types import OAPI
from drakaina.types import ORPC
from drakaina.utils import unwrap_func

__all__ = ("JsonRPCv2",)

log = getLogger()

DEFAULT_OPENRPC_SCHEMA = ORPC.OpenRPC(
    openrpc=ORPC.SUPPORTED_VERSION,
    info=ORPC.Info(version="1.0.0", title="JSON-RPC 2.0 service"),
    servers=[ORPC.Server(name="main", url="/")],
    methods=[],
)
"""This is the default OpenRPC schema.

Specify an explicit schema template
'openrpc_schema_template' to control it.
"""

DEFAULT_OPENAPI_SCHEMA = OAPI.OpenAPI(
    openapi=OAPI.SUPPORTED_VERSION,
    info=OAPI.Info(version="1.0.0", title="JSON-RPC 2.0 service"),
    servers=[OAPI.Server(url="/")],
    paths=OAPI.Paths(),
)
"""This is the default OpenAPI schema.

Specify an explicit schema template
'openapi_schema_template' to control it.
"""


# JSON-RPC Error classes


class JsonRPCError(RPCError):
    """JSON-RPC Common error

    Reserved for implementation-defined server-errors.
    Codes -32000 to -32099.

    """

    code: int = -32000
    default_message: str = "Server error"
    id: int | str = None
    data: Any = None

    def __init__(
        self,
        *args,
        message: str = None,
        id: int | str = None,
        data: Any | None = None,
    ):
        super().__init__(*args)

        self.id = id
        if message:
            self.message = message

        if self.message and data:
            self.data = {"text": self.message.strip(), "details": data}
        elif self.message:
            self.data = self.message.strip()
        elif data:
            self.data = data

    def __str__(self) -> str:
        return f"{self.__class__.__name__} ({self.code} {self.default_message})"

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} ({self.code} {self.default_message})>"
        )

    def as_dict(self) -> JSONRPCResponseObject:
        error = dict(
            jsonrpc="2.0",
            error={"code": self.code, "message": self.default_message},
            id=self.id,
        )

        if self.data:
            error["error"]["data"] = self.data

        return error


class InvalidRequestError(JsonRPCError):
    """Invalid Request

    The JSON sent is not a valid Request object.

    """

    code = -32600
    default_message = "Invalid Request"


class MethodNotFoundError(JsonRPCError):
    """Method not found

    The method does not exist / is not available.

    """

    code = -32601
    default_message = "Method not found"


class InvalidParamsError(JsonRPCError):
    """Invalid params

    Invalid method parameter(s).

    """

    code = -32602
    default_message = "Invalid params"


class InternalError(JsonRPCError):
    """Internal error

    Internal JSON-RPC error.

    """

    code = -32603
    default_message = "Internal error"


class ParseError(JsonRPCError):
    """Parse error

    Invalid JSON was received by the server.
    An error occurred on the server while parsing the JSON text.

    """

    code = -32700
    default_message = "Parse error"


# Implementation of Drakaina errors


class AuthenticationFailedJRPCError(JsonRPCError):
    """Authentication failed"""

    code = -32010
    default_message = "Authentication failed"


class InvalidTokenJRPCError(AuthenticationFailedJRPCError):
    """Invalid token error"""

    code = -32011
    default_message = "Invalid token"


class ForbiddenJRPCError(AuthenticationFailedJRPCError):
    """Forbidden error"""

    code = -32012
    default_message = "Forbidden"


class InvalidPermissionsJRPCError(ForbiddenJRPCError):
    """Invalid permissions error"""

    code = -32013
    default_message = "Invalid permissions"


# JSON-RPC v2.0 implementation


class JsonRPCv2(BaseRPCProtocol):
    """JSON-RPC 2.0 implementation.

    :param registry:
        Registry of remote procedures.
        Default: `drakaina.registries.rpc_registry` (generic module instance)
    :param serializer:
        Serializer object. Default: `JsonSerializer` (stdlib.json)
    :param schema_serializer:
        The serializer object to serialize the schema.
        Default: `JsonSerializer` (stdlib.json)
    :param openrpc:
        An object (TypedDict) of OpenRPC containing general information,
        information about the server(s). It must not contain a method schema.
    :param openapi:
        An object (TypedDict) of OpenAPI containing general information,
        information about the server(s). It must not contain a paths schema.

    """

    __slots__ = (
        "allow_service_discovery",
        "schema_url",
        "openrpc_schema_template",
        "openapi_schema_template",
        "__openrpc_schema",
        "__openapi_schema",
    )

    base_error_class: ClassVar = JsonRPCError
    """Base class for JSON-RPC 2.0 protocol implementation."""

    default_error_class: ClassVar = InternalError
    """Default JSON-RPC 2.0 error class to represent internal exceptions."""

    _errors_map: ClassVar = {
        Exception: JsonRPCError,
        RPCError: JsonRPCError,
        BadRequestError: InvalidRequestError,
        NotFoundError: MethodNotFoundError,
        InvalidParametersError: InvalidParamsError,
        InternalServerError: InternalError,
        SerializationError: InternalError,
        DeserializationError: ParseError,
        AuthenticationFailedError: AuthenticationFailedJRPCError,
        InvalidTokenError: InvalidTokenJRPCError,
        ForbiddenError: ForbiddenJRPCError,
        InvalidPermissionsError: InvalidPermissionsJRPCError,
    }

    def __init__(
        self,
        registry: RPCRegistry | None = None,
        serializer: BaseSerializer | None = None,
        schema_serializer: BaseSerializer | None = None,
        openrpc: ORPC.OpenRPC | None = None,
        openapi: OAPI.OpenAPI | None = None,
        allow_service_discovery: bool | None = None,
        schema_url: str | None = None,
    ):
        if serializer is None:
            serializer = JsonSerializer()

        super().__init__(
            registry=registry,
            serializer=serializer,
            schema_serializer=schema_serializer,
        )
        self.allow_service_discovery = allow_service_discovery
        self.schema_url = schema_url

        # Schemas
        self.openrpc_schema_template = (
            openrpc if openrpc is not None else DEFAULT_OPENRPC_SCHEMA
        )
        self.__openrpc_schema = None
        self.openapi_schema_template = (
            openapi if openapi is not None else DEFAULT_OPENAPI_SCHEMA
        )
        self.__openapi_schema = None

        if allow_service_discovery:
            self.registry.register_rpc_extension(
                extension_procedure=partial(rpc_discover, protocol=self),
                extension_name="rpc.discover",
            )

    def handle(
        self,
        rpc_request: JSONRPCRequest,
        request: Any | None = None,
    ) -> JSONRPCResponse | None:
        """Handles a procedure call or batch of procedure call

        :param rpc_request:
            RPC request in protocol format.
        :type rpc_request: JSONRPCRequest
        :param request:
            Optional parameter that can be passed as an
            argument to the procedure.
        :type request: Any
        :return:
            Returns the result in protocol format.
        :rtype: JSONRPCResponse

        """
        # Check bad request
        if not (isinstance(rpc_request, (dict, list)) and len(rpc_request) > 0):
            return InvalidRequestError().as_dict()

        # Handle batch request
        if isinstance(rpc_request, list):
            batch_result = []
            for request_object in rpc_request:
                result = self.execute(request_object, request=request)
                if result is not None:
                    batch_result.append(result)
            if len(batch_result) == 0:
                return None
            return batch_result

        # Handle single request
        return self.execute(rpc_request, request=request)

    def execute(
        self,
        procedure_call: JSONRPCRequestObject,
        request: Any | None = None,
    ) -> JSONRPCResponseObject | None:
        """Execute a remote procedure call.

        :param procedure_call:
            RPC request object in protocol format.
        :type procedure_call: JSONRPCRequestObject
        :param request:
            Optional parameter that can be passed as an
            argument to the procedure. By default, None will be passed.
        :type request: Any
        :return:
            Returns a result object in protocol format.
        :rtype: JSONRPCResponseObject

        """
        if not isinstance(procedure_call, dict):
            return InvalidRequestError().as_dict()
        method: str = procedure_call.get("method")
        params: list | dict | None = procedure_call.get("params")
        request_id: int | str | None = procedure_call.get("id")

        # Validate protocol
        if (
            procedure_call.get("jsonrpc") != "2.0"
            or not isinstance(method, str)
            or not (params is None or isinstance(params, (list, dict)))
            or not (request_id is None or isinstance(request_id, (int, str)))
        ):
            return InvalidRequestError(id=request_id).as_dict()

        # Getting procedure
        procedure = self.registry[method]
        if procedure is None:
            if request_id is None:
                return None
            return MethodNotFoundError(id=request_id).as_dict()

        # Prepare parameters
        args, kwargs = (), {}
        if params is not None:
            if isinstance(params, list):
                args = (request, *params)
            elif isinstance(params, dict):
                kwargs = dict(request=request, **params)
        else:
            kwargs = dict(request=request)

        # Execute RPC method
        try:
            result = procedure(*args, **kwargs)
        except RPCError as err:
            error = self.handle_error(err)
            error.id = request_id
            return error.as_dict()
        except Exception as err:
            return InternalError(message=str(err), id=request_id).as_dict()

        if request_id is None:
            return None

        return dict(jsonrpc="2.0", result=result, id=request_id)

    def openrpc_schema(self) -> ORPC.OpenRPC:
        """Implementation of OpenRPC specification.

        https://spec.open-rpc.org/

        """
        if not self.__openrpc_schema:
            openrpc: ORPC.OpenRPC = deepcopy(self.openrpc_schema_template)
            if not isinstance(openrpc.get("methods"), list):
                openrpc["methods"] = []

            for method_name, procedure_ in self.registry.items():
                procedure = unwrap_func(procedure_)
                try:
                    # TODO
                    method_metadata: dict = getattr(procedure, RPC_META)
                    method_schema: MethodSchema = getattr(procedure, RPC_SCHEMA)
                except AttributeError:
                    # TODO handle error
                    err = (
                        f"AttributeError;\n{getmembers(procedure_)=};"
                        f"\n{getmembers(procedure)=};",
                    )
                    print(f"openrpc_schema; {err}")
                    log.error(err)
                    continue

                openrpc["methods"].append(
                    ORPC.method_by_schema(
                        method_schema,
                        default_name=method_name,
                    ),
                )

            self.__openrpc_schema = openrpc
        return self.__openrpc_schema

    def openapi_schema(self) -> dict:
        """Implementation of OpenAPI specification.

        https://spec.openapis.org/oas/latest.html

        """
        if not self.__openapi_schema:
            openapi: OAPI.OpenAPI = deepcopy(self.openapi_schema_template)

            for method_name, procedure_ in self.registry.items():
                procedure = unwrap_func(procedure_)
                try:
                    # TODO
                    method_metadata: dict = getattr(procedure, RPC_META)
                    method_schema: MethodSchema = getattr(procedure, RPC_SCHEMA)
                except AttributeError:
                    err = (  # TODO handle error
                        f"AttributeError;\n{getmembers(procedure_)=};"
                        f"\n{getmembers(procedure)=};",
                    )
                    print(f"openrpc_schema; {err}")
                    log.error(err)
                    continue

                path_name = self.schema_url + "#" + method_name
                operation = OAPI.operation_by_schema(
                    method_schema,
                    default_name=method_name,
                )
                openapi["paths"][path_name] = OAPI.PathItem(post=operation)

            self.__openapi_schema = openapi
        return self.__openapi_schema

    def _validate_rpc_url(self):
        assert bool(self.schema_url), (
            "To use service description schemes, "
            "you must specify the URL responsible for processing "
            "RPC requests."
        )


def rpc_discover(protocol: JsonRPCv2) -> dict[str, Any]:
    """Returns an OpenRPC schema as a description of this service.

    :return: OpenRPC Schema.

    """
    if protocol.schema_url is not None:
        protocol._validate_rpc_url()  # noqa
        schema = {"$ref": protocol.schema_url}
    else:
        schema = protocol.openrpc_schema()

    return schema



from __future__ import annotations

import re
from functools import partial
from logging import Logger
from sys import version_info
from typing import Iterable

from drakaina import ENV_APP
from drakaina.exceptions import BadRequestError
from drakaina.middleware.base import BaseMiddleware
from drakaina.middleware.exception import ExceptionMiddleware
from drakaina.middleware.request_wrapper import RequestWrapperMiddleware
from drakaina.rpc_protocols import BaseRPCProtocol
from drakaina.rpc_protocols import JsonRPCv2
from drakaina.types import WSGIApplication
from drakaina.types import WSGIEnvironment
from drakaina.types import WSGIInputStream
from drakaina.types import WSGIResponse
from drakaina.types import WSGIStartResponse
from drakaina.utils import match_path

if version_info < (3, 9):
    from typing import Type
    from typing import Union
    from typing_extensions import TypeAlias

    Middlewares: TypeAlias = Iterable[
        Union[Type[BaseMiddleware], partial[BaseMiddleware]],
    ]
elif version_info < (3, 10):
    from typing import Union
    from typing_extensions import TypeAlias

    Middlewares: TypeAlias = Iterable[
        Union[type[BaseMiddleware], partial[BaseMiddleware]],
    ]
else:
    from typing import TypeAlias

    Middlewares: TypeAlias = Iterable[
        type[BaseMiddleware] | partial[BaseMiddleware],
    ]

__all__ = ("WSGIHandler",)

ALLOWED_METHODS = ("OPTIONS", "GET", "POST")


class WSGIHandler:
    """Implementation of WSGI protocol.

    # todo: write about urls (route, provide_ args)

    :param route:
    :type route: str | re.Pattern
    :param handler: RPC protocol implementation.
    :type handler: BaseRPCProtocol
    :param middlewares: List of WSGI middlewares.
    :type middlewares: Iterable[type[BaseMiddleware] | partial[BaseMiddleware]]
    :param logger: A `logging.Logger` object.
    :type logger: Logger
    :param max_content_size: Limiting request body size for DoS protection.
    :type max_content_size: int
    :param openrpc_url: OpenRPC Schema URL.
        The recommended document name is `openrpc.json`.
    :type openrpc_url: str | re.Pattern
    :param openapi_url: OpenAPI Schema URL.
        The recommended document name is `openapi.json`.
    :type openapi_url: str | re.Pattern

    """

    __slots__ = (
        "environ",
        "start_response",
        "handler",
        "route",
        "logger",
        "max_content_size",
        "openrpc_url",
        "openapi_url",
        "_rpc_content_type",
        "_allowed_methods",
        "_middlewares_chain",
    )

    environ: WSGIEnvironment
    start_response: WSGIStartResponse
    _middlewares_chain: WSGIApplication | BaseMiddleware

    def __init__(
        self,
        route: str | re.Pattern | None = None,
        handler: BaseRPCProtocol | None = None,
        middlewares: Middlewares | None = None,
        logger: Logger | None = None,
        max_content_size: int = 4096,
        openrpc_url: str | re.Pattern | None = None,
        openapi_url: str | re.Pattern | None = None,
    ):
        self.handler = handler if handler is not None else JsonRPCv2()
        self._rpc_content_type = self.handler.content_type

        self.route = route
        if isinstance(self.route, str) and not self.route.startswith("/"):
            # need it?
            self.route = "/" + self.route
        schema_url = getattr(self.handler, "schema_url", None)
        if (
            self.route is not None
            and hasattr(self.handler, "schema_url")
            and schema_url is None
        ):
            setattr(self.handler, "schema_url", self.route)

        self.logger = logger
        self.max_content_size = int(max_content_size)
        self.openrpc_url = openrpc_url
        self.openapi_url = openapi_url

        if openrpc_url or openapi_url:
            self._allowed_methods = ", ".join(ALLOWED_METHODS)
        else:
            self._allowed_methods = ", ".join(
                [m for m in ALLOWED_METHODS if m != "GET"],
            )

        # Build middleware stack
        self._middlewares_chain = self._wsgi_app
        kw = {"is_async": False}
        for mw in reversed(middlewares or ()):
            if (
                isinstance(mw, partial) and issubclass(mw.func, BaseMiddleware)
            ) or issubclass(mw, BaseMiddleware):
                self._middlewares_chain = mw(self._middlewares_chain, **kw)
            self._middlewares_chain = mw(self._middlewares_chain)

        # The middleware for handling exceptions in the middleware according
        #  to the RPC protocol.
        self._middlewares_chain = ExceptionMiddleware(
            RequestWrapperMiddleware(self._middlewares_chain, **kw),  # noqa
            handler=self.handler,
            logger=self.logger,
            **kw,
        )

    def __call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        environ[ENV_APP] = self
        return self._middlewares_chain(environ, start_response)  # noqa

    def _wsgi_app(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        self.environ = environ
        self.start_response = start_response

        method = environ["REQUEST_METHOD"]
        if method in ALLOWED_METHODS:
            return getattr(self, method.lower())()

        return self._method_not_allowed()

    def get(self) -> WSGIResponse:
        if match_path(self.openrpc_url, self.environ["PATH_INFO"]):
            response_body = self.handler.get_raw_openrpc_schema()
            response_headers = [
                ("Content-Type", self.handler.schema_content_type),
                ("Content-Length", str(len(response_body))),
            ]
        elif match_path(self.openapi_url, self.environ["PATH_INFO"]):
            response_body = self.handler.get_raw_openapi_schema()
            response_headers = [
                ("Content-Type", self.handler.schema_content_type),
                ("Content-Length", str(len(response_body))),
            ]
        else:
            return self._not_found()

        self.start_response("200 OK", response_headers)
        return (response_body,)

    def post(self) -> WSGIResponse:
        if not match_path(self.route, self.environ["PATH_INFO"]):
            return self._not_found()

        wsgi_input: WSGIInputStream = self.environ["wsgi.input"]

        content_type = self.environ.get("CONTENT_TYPE")
        content_length = int(self.environ.get("CONTENT_LENGTH") or 0)
        if (
            not (content_type and content_length)
            or content_type != self._rpc_content_type
            or content_length > self.max_content_size
        ):
            if content_type != self._rpc_content_type:
                response_status = "415 Unsupported Media Type"
            else:
                response_status = "400 Bad Request"
            # Return RPC error
            response_body = self.handler.get_raw_error(BadRequestError())
        else:
            response_status = "200 OK"
            response_body = self.handler.handle_raw_request(
                wsgi_input.read(content_length),
                request=self.environ,
            )

        response_headers = [
            ("Content-Type", self._rpc_content_type),
            ("Content-Length", str(len(response_body))),
        ]
        env_response_headers = self.environ.get("response", {}).get("headers")
        if isinstance(env_response_headers, list):
            response_headers.extend(env_response_headers)

        self.start_response(response_status, response_headers)

        return (response_body,)

    def options(self) -> WSGIResponse:
        response_headers = [
            ("Allow", self._allowed_methods),
            ("Content-Length", "0"),
        ]
        self.start_response("200 OK", response_headers)
        yield b""

    def _not_found(self) -> WSGIResponse:
        response_headers = []
        self.start_response("404 Not Found", response_headers)
        yield b""

    def _method_not_allowed(self) -> WSGIResponse:
        response_headers = [("Allow", self._allowed_methods)]
        self.start_response("405 Method Not Allowed", response_headers)
        yield b""



from __future__ import annotations

from typing import Iterable
from typing import Optional

from drakaina.middleware.base import BaseMiddleware
from drakaina.types import ASGIReceive
from drakaina.types import ASGIScope
from drakaina.types import ASGISend
from drakaina.types import WSGIApplication
from drakaina.types import WSGIEnvironment
from drakaina.types import WSGIResponse
from drakaina.types import WSGIStartResponse
from drakaina.utils import iterable_str_arg

__all__ = ("CORSMiddleware",)

DEFAULT_HEADERS = (
    "Accept, Accept-Encoding, Accept-Language, Authorization, "
    "Content-Language, Content-Type, DNT, Origin, User-Agent, X-Requested-With"
)
ALL_METHODS = "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT"
DEFAULT_METHODS = "GET, POST, OPTIONS"


class CORSMiddleware(BaseMiddleware):
    """Middleware for providing CORS headers and handling preflight requests.

    See: https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS

    :param allow_origin:
        The "Access-Control-Allow-Origin" header.
    :type allow_origin: Union[str, Iterable[str]]
    :param allow_methods:
        The "Access-Control-Allow-Methods" header.
        Default: "GET, POST, OPTIONS".
    :type allow_methods: Union[str, Iterable[str]]
    :param allow_headers:
        The "Access-Control-Allow-Headers" header.
        Default: "Accept, Accept-Encoding, Authorization,
        Content-Type, DNT, Origin, User-Agent, X-Requested-With".
    :type allow_headers: Union[str, Iterable[str]]
    :param allow_credentials:
        The "Access-Control-Allow-Credentials" header. Default: None.
    :type allow_credentials: bool
    :param expose_headers:
        The "Access-Control-Expose-Headers" header. Default: None.
    :type expose_headers: Union[str, Iterable[str]]
    :param max_age:
        The "Access-Control-Max-Age" header. Default: 3600 sec. (1 hour).
    :type max_age: Union[int, str]

    """

    __slots__ = (
        "_cors_headers",
        "_preflight_cors_headers",
        "_allow_origin",
        "_allow_credentials",
        "_allow_headers",
    )

    def __init__(
        self,
        app: WSGIApplication,
        allow_origin: str | Iterable[str],
        allow_methods: Optional[str | Iterable[str]] = None,
        allow_headers: Optional[str | Iterable[str]] = None,
        allow_credentials: Optional[bool] = None,
        expose_headers: Optional[str | Iterable[str]] = None,
        max_age: int = 3600,
        **kwargs,
    ):
        super().__init__(app, **kwargs)

        self._allow_origin = ", ".join(iterable_str_arg(allow_origin))

        if allow_methods:
            if "*" in allow_methods:
                _allow_methods = ALL_METHODS
            else:
                _allow_methods = ", ".join(iterable_str_arg(allow_methods))
        else:
            _allow_methods = DEFAULT_METHODS

        if allow_headers:
            self._allow_headers = ", ".join(
                sorted(
                    set(iterable_str_arg(DEFAULT_HEADERS))
                    | set(iterable_str_arg(allow_headers)),
                ),
            )
        else:
            self._allow_headers = DEFAULT_HEADERS

        if expose_headers:
            _expose_headers = ", ".join(iterable_str_arg(expose_headers))
        else:
            _expose_headers = None

        self._allow_credentials = allow_credentials

        # Prepare CORS response headers
        self._cors_headers = []
        if self._allow_credentials:
            self._cors_headers.append(
                ("Access-Control-Allow-Credentials", "true"),
            )
        if _expose_headers:
            self._cors_headers.append(
                ("Access-Control-Expose-Headers", _expose_headers),
            )

        # Prepare pre-flight CORS response headers
        self._preflight_cors_headers = [
            ("Access-Control-Allow-Methods", _allow_methods),
            ("Access-Control-Max-Age", str(max_age)),
            ("Content-Length", "0"),
        ]
        if "*" not in self._allow_origin or self._allow_credentials:
            self._preflight_cors_headers.append(("Vary", "Origin"))

        if self._allow_credentials:
            self._preflight_cors_headers.append(
                ("Access-Control-Allow-Credentials", "true"),
            )

    def __wsgi_call__(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        if environ.get("HTTP_ORIGIN"):
            if environ.get("REQUEST_METHOD") == "OPTIONS":
                return self.options(environ, start_response)
            return self.app(
                environ,
                self._start_response_with_cors(environ, start_response),
            )
        else:
            return self.app(environ, start_response)

    async def __asgi_call__(
        self,
        scope: ASGIScope,
        receive: ASGIReceive,
        send: ASGISend,
    ):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
        ...
        # if scope.get("HTTP_ORIGIN"):
        #     if scope.get("REQUEST_METHOD") == "OPTIONS":
        #         return self.options(environ, start_response)
        #     return await self.app(
        #         environ,
        #         self._add_cors_headers(environ, start_response),
        #     )
        # else:
        #     return await self.app(environ, start_response)

    def options(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIResponse:
        """Response for OPTIONS request"""
        request_origin = environ["HTTP_ORIGIN"]
        request_headers = environ.get("HTTP_ACCESS_CONTROL_REQUEST_HEADERS")

        response_headers = self._preflight_cors_headers[:]

        if "*" in self._allow_origin or request_origin in self._allow_origin:
            if "*" not in self._allow_origin or self._allow_credentials:
                response_headers.append(
                    ("Access-Control-Allow-Origin", request_origin),
                )
            else:
                response_headers.append(("Access-Control-Allow-Origin", "*"))

        if "*" in self._allow_headers and request_headers is not None:
            response_headers.append(
                ("Access-Control-Allow-Headers", request_headers),
            )
        if "*" not in self._allow_headers:
            response_headers.append(
                ("Access-Control-Allow-Headers", self._allow_headers),
            )

        start_response("200 OK", response_headers)
        yield b""

    def _start_response_with_cors(
        self,
        environ: WSGIEnvironment,
        start_response: WSGIStartResponse,
    ) -> WSGIStartResponse:
        """Wraps the start_response method, and includes the CORS header
        for the specified origin.
        """
        request_origin = environ["HTTP_ORIGIN"]

        cors_headers = self._cors_headers[:]

        if "*" in self._allow_origin:
            if "HTTP_COOKIE" in environ:
                cors_headers.append(
                    ("Access-Control-Allow-Origin", request_origin),
                )
                cors_headers.append(("Vary", "Origin"))
            else:
                cors_headers.append(("Access-Control-Allow-Origin", "*"))
        elif (
            "*" not in self._allow_origin
            and request_origin in self._allow_origin
        ):
            cors_headers.append(("Access-Control-Allow-Origin", request_origin))
            cors_headers.append(("Vary", "Origin"))

        def response_with_cors(status, headers, exc_info=None):
            headers.extend(cors_headers)
            return start_response(status, headers, exc_info)

        return response_with_cors
